# flake8: noqa

"""Finesse legacy (Finesse 2) kat-script parser."""

from collections import OrderedDict
import logging
import re

from sly import Lexer, Parser

from .. import Model, analysis, components, constants, detectors, symbols, gaussian
from ..frequency import Frequency
from ..utilities import opened_file

LOGGER = logging.getLogger(__name__)


def select_powerdetector(*args, **kwargs):
    if len(args) + len(kwargs) <= 2:
        return detectors.PowerDetector(*args, **kwargs)
    elif len(args) + len(kwargs) <= 4:
        new_kwargs = {}
        for k, v in kwargs.items():
            if k == "f1" or k == "phase1":
                new_kwargs[k.strip("1")] = v
            else:
                new_kwargs[k] = v
        return detectors.PowerDetectorDemod1(*args, **new_kwargs)
    elif len(args) + len(kwargs) <= 6:
        return detectors.PowerDetectorDemod2(*args, **kwargs)


def select_quantum_noisedetector(*args, **kwargs):
    shot_only = kwargs.pop("shot_only")
    if len(args) + len(kwargs) <= 3:
        if shot_only:
            return detectors.QuantumShotNoiseDetector(*args, **kwargs)
        else:
            return detectors.QuantumNoiseDetector(*args, **kwargs)
    elif len(args) + len(kwargs) <= 5:
        new_kwargs = {}
        for k, v in kwargs.items():
            if k == "f1" or k == "phase1":
                new_kwargs[k.strip("1")] = v
            else:
                new_kwargs[k] = v
        if shot_only:
            return detectors.QuantumShotNoiseDetectorDemod1(*args, **new_kwargs)
        else:
            return detectors.QuantumNoiseDetectorDemod1(*args, **new_kwargs)
    elif len(args) + len(kwargs) <= 7:
        if shot_only:
            return detectors.QuantumShotNoiseDetectorDemod2(*args, **kwargs)
        else:
            return detectors.QuantumNoiseDetectorDemod2(*args, **kwargs)


def get_const(consts, value):
    if type(value) is str and value.strip("+-") in consts:
        ret = consts[value.strip("+-")]
        if value.startswith("-"):
            ret *= -1
    else:
        ret = value
    return ret


def get_model_element(model, element):
    try:
        return model.elements[element]
    except KeyError:
        raise KeyError(f"Element '{element}' not in model.")


class KatParser:
    """Kat file lexer, parser and builder."""

    def parse(self, text, model=None, **kwargs):
        """Parse kat code into a model.

        Parameters
        ----------
        text : str
            String containing the kat code to be parsed.

        model : :class:`.Model`, optional
            Model object to add components to. If not specified, a new model will be
            created.

        ignored_blocks : list, optional
            A list of names of ``FTBLOCK`` sections in the kat code to leave out of the
            model; defaults to empty list.

        Returns
        -------
        :class:`.Model`
            The constructed model.

        Raises
        ------
        :class:`.KatParserError`
            If an error occurs during parsing or building.
        """
        if model is not None:
            LOGGER.info(f"Parsing into existing model {model!r}.")
        else:
            model = Model()

        return self._build(self._parse(text), model, **kwargs)

    def parse_file(self, path, model=None, **kwargs):
        """Parse kat code from a file into a model.

        Parameters
        ----------
        path : str or :py:class:`io.FileIO`
            The path or file object to read kat script from. If an open file object is
            passed, it will be read from and left open. If a path is passed, it will be
            opened, read from, then closed.

        model : :class:`.Model`, optional
            Model object to add components to. If not specified, a new model will be
            created.

        Other Parameters
        ----------------
        **kwargs
            Keyword parameters supported by :meth:`.parse`.

        Raises
        ------
        :class:`.KatParserError`
            If an error occurs during parsing or building.
        """
        with opened_file(path, "r") as fobj:
            LOGGER.info(f"Parsing kat script from {fobj.name}")
            return self.parse(fobj.read(), model=model, **kwargs)

    def _parse(self, text):
        """Parses kat code into a model.

        Parameters
        ----------
        text : str
            String containing the kat code to be parsed.

        Returns
        -------
        parser : :class:`.sly.Parser`
            The parser object containing the parsed blocks.

        Raises
        ------
        KatParserError
            If an error occurs during parsing.
        """
        lexer = _KatLEX()
        parser = _KatYACC()

        # As we have no way of detecting EOF and calling pop_state() from
        # within the component lexer, we must ensure that all files end in a
        # newline
        text = f"{text}\n"

        # Trim any whitespace within $$ strings
        matches = re.findall(r"\$\$[^$]+\$\$", text)
        for match in matches:
            text = text.replace(match, re.sub(r"\s+", "", match))

        tokens = lexer.tokenize(text)
        parser.parse(tokens)

        errors = sorted(lexer.errors + parser.errors, key=lambda tup: tup[1])
        if len(errors) > 0:
            raise KatParserError(errors, text)

        for warning in lexer.warnings:
            LOGGER.warn(f"{warning[1]}:{find_column(text, warning[2])}: {warning[0]}")

        return parser

    def _build(self, parser, model=None, ignored_blocks=None):
        """Constructs a new model or appends to an existing model using the parsed kat
        code.

        Parameters
        ----------
        parser : :class:`.sly.Parser`
            The parser object containing the parsed blocks.

        model : :class:`.Model`
            Model object to update.

        ignored_blocks : list, optional
            A list of names of ``FTBLOCK`` sections in the kat code to leave out of the model;
            defaults to empty list.

        Returns
        -------
        :class:`.Model`
            The constructed model.
        """

        blocks = parser.blocks

        if ignored_blocks is None:
            ignored_blocks = []

        def parse_parameter(param):
            if type(param) is not str or "$" not in param:
                return param
            local = dict(map(lambda f: (f.name, f), model.frequencies))
            local.update(model.alternate_name_map)
            local = {**local, **model.elements}
            if param.endswith("$"):
                p = re.sub("([a-zA-Z_][a-zA-Z0-9_:.]*)", r"\1.ref", param)
            else:
                p = param
            return eval(f"{p}".replace("$", ""), local)

        LOGGER.info("Building model")

        component_constructors = {
            "lasers": components.Laser,
            "squeezers": components.Squeezer,
            "mirrors": components.Mirror,
            "beamsplitters": components.Beamsplitter,
            "directional_beamsplitters": components.DirectionalBeamsplitter,
            "isolators": components.Isolator,
            "modulators": components.Modulator,
            "lenses": components.Lens,
        }

        detector_constructors = {
            "amplitude_detectors": detectors.AmplitudeDetector,
            "beam_detectors": (detectors.CCDPixel, detectors.FieldPixel),
            "beam_property_detectors": detectors.BeamPropertyDetector,
            "power_detectors": select_powerdetector,
            "quantum_noise_detectors": select_quantum_noisedetector,
        }
        model = model or Model()
        node_names = ["node", "node1", "node2", "node3", "node4"]
        nodes = {}
        consts = {}

        ignored_blocks = set(ignored_blocks)

        if ignored_blocks:
            ignored_block_list = ", ".join(ignored_blocks)
            LOGGER.debug(f"Ignoring blocks {ignored_block_list}.")

        for key in ignored_blocks:
            # TODO: make KatParserError behave like the kat3 parser's, i.e. it can be instantiated
            # with a single message and not just a list of errors, then if a KeyError is thrown
            # here, rethrow a KatParserError with a message about the ignored block not existing.
            blocks.pop(key)

        # First pass, just grab constants
        for block, d in blocks.items():
            for k, v in d["constants"].items():
                consts[k] = v

        def apply_constants(comp):
            if isinstance(comp, list):
                for k, v in enumerate(comp):
                    if isinstance(v, dict) or isinstance(v, list):
                        apply_constants(v)
                    else:
                        comp[k] = get_const(consts, v)
            if isinstance(comp, dict):
                for k, v in comp.items():
                    if isinstance(v, dict) or isinstance(v, list):
                        apply_constants(v)
                    else:
                        comp[k] = get_const(consts, v)

        # And apply any constant renaming
        for block, d in blocks.items():
            for name, comp_type in d.items():
                if name == "constants":
                    continue
                apply_constants(comp_type)

        # Variables
        for block, d in blocks.items():
            for k, v in d["variables"].items():
                model.add(components.Variable(k, v))

        # Next grab source frequencies
        for block, d in blocks.items():
            for f in d["frequencies"]:
                model.add_frequency(
                    Frequency(f["name"], model, symbols.Constant(f["f"]))
                )

        # Next construct all frequencies & normal components,
        # and get node names
        for block, d in blocks.items():
            for name, constructor in component_constructors.items():
                for comp in d[name]:
                    args = []
                    ns = []
                    lineno = None
                    for k, v in comp.items():
                        if k == "lineno":
                            lineno = v
                        elif k in node_names:
                            ns.append(v)
                        else:
                            args.append(v)
                    el = constructor(*args)
                    if lineno:
                        el._legacy_script_line_number = lineno
                    model.add(el)
                    for i, n in enumerate(ns):
                        if n in nodes and n != "dump":
                            raise ValueError(
                                f"In block '{block}': {args[0]}: "
                                f"Node '{n}' already assigned to "
                                f"'{nodes[n][0]}'."
                            )
                        else:
                            nodes[n] = (args[0], ns.index(n) + 1)
                            # get the output node object corresponding to
                            # this node and tag it
                            _ni = getattr(
                                getattr(
                                    get_model_element(model, comp["name"]), f"p{i + 1}"
                                ),
                                "o",
                            )
                            try:
                                model.tag_node(_ni, n)
                            except:
                                pass

        # Connect all of the components up with spaces
        for block, d in blocks.items():
            for space in d["spaces"]:
                try:
                    comp1 = nodes[space["node1"]]
                    node1 = getattr(get_model_element(model, comp1[0]), f"p{comp1[1]}")
                except KeyError:
                    name = space["name"] + "_" + space["node1"]
                    comp1 = components.Nothing(name)
                    model.add(comp1)
                    node1 = comp1.p1
                    nodes[space["node1"]] = (name, 1)
                try:
                    comp2 = nodes[space["node2"]]
                    node2 = getattr(get_model_element(model, comp2[0]), f"p{comp2[1]}")
                except KeyError:
                    name = space["name"] + "_" + space["node2"]
                    comp2 = components.Nothing(name)
                    model.add(comp2)
                    node2 = comp2.p1
                    nodes[space["node2"]] = (name, 1)
                kwargs = {}
                lineno = None
                for k, v in space.items():
                    if k == "lineno":
                        lineno = v
                    elif k not in node_names:
                        kwargs[k] = v

                el = components.Space(**kwargs, portA=node1, portB=node2)
                model.add(el)
                if lineno:
                    el._legacy_script_line_number = lineno

        # Create fsigs
        for block, d in blocks.items():
            for fsig in d["fsigs"]:
                if model.fsig.f.value is None:
                    model.fsig.f = parse_parameter(fsig["f"])
                    model.alternate_name_map["fs"] = model.fsig.f.ref
                    model.alternate_name_map["mfs"] = model.fsig.f.ref
                elif model.fsig.f.value != parse_parameter(fsig["f"]):
                    raise ValueError("Cannot have more than one signal frequency.")

                # fsig command is specifying an input not just a frequency
                if len(fsig.keys()) == 2:
                    # Need to make some dummy object here to reference
                    # the model fsig value legacy issues of dealing with
                    # a single parameter that can have multiple names
                    model.fsig.f = fsig["f"]
                    model.alternate_name_map[fsig["name"]] = model.fsig
                else:
                    comp = get_model_element(model, fsig["component"])
                    mod_type = fsig["mod_type"]
                    scaling = 1
                    if isinstance(comp, components.Mirror) or isinstance(
                        comp, components.Beamsplitter
                    ):
                        node = comp.mech.z
                        if mod_type is None or mod_type == "phase":
                            scaling = model.lambda0 / (2 * constants.PI)
                        else:
                            raise ValueError(
                                f"Unsupported signal type '{mod_type}' at a mirror."
                            )
                    else:
                        if mod_type == "amp":
                            node = comp.amp.i
                        elif mod_type == "phase":
                            node = comp.phs.i
                        elif mod_type == "freq":
                            node = comp.frq.i
                        elif mod_type is None:
                            if isinstance(comp, components.Laser):
                                node = comp.frq.i
                            elif isinstance(comp, components.Space):
                                node = comp.h.i
                            else:
                                node = comp.phs.i
                        else:
                            raise ValueError(f"Unknown signal type '{mod_type}'.")
                    signal = components.SignalGenerator(
                        fsig["name"], node, fsig["amp"] * scaling, fsig["phase"]
                    )
                    model.add(signal)

        # Create detectors
        for block, d in blocks.items():
            for name, constructor in detector_constructors.items():
                for det in d[name]:
                    args = {}
                    lineno = None
                    for k, v in det.items():
                        if k == "lineno":
                            lineno = v
                        elif k not in node_names:
                            if isinstance(v, list):
                                args[k] = [parse_parameter(x) for x in v]
                            else:
                                args[k] = parse_parameter(v)
                        else:
                            direction = "o"
                            if v.endswith("*"):
                                v = v.strip("*")
                                direction = "i"
                            comp = get_model_element(model, nodes[v][0])
                            if isinstance(comp, components.Nothing) or isinstance(
                                comp, components.Laser
                            ):
                                if direction == "i":
                                    direction = "o"
                                else:
                                    direction = "i"
                            node = f"p{nodes[v][1]}.{direction}"
                            n = comp
                            for attr in node.split("."):
                                n = getattr(n, attr)
                            args["node"] = n

                    if name == "beam_detectors":
                        if args["f"] is None:  # CCDPixel
                            del args["f"]
                            constructor = constructor[0]
                        else:  # FieldPixel
                            constructor = constructor[1]

                    comp = constructor(**args)
                    comp._legacy_script_line_number = lineno
                    model.add(comp)

        # Gouy detector
        for block, d in blocks.items():
            for det in d["gouy"]:
                args = [det["name"]]
                for space in det["space_list"]:
                    args.append(get_model_element(model, space))
                el = detectors.Gouy(*args, direction=det["direction"])
                el._legacy_script_line_number = det["lineno"]
                model.add(el)

        # Motion detector
        for block, d in blocks.items():
            for det in d["motion_detectors"]:
                name = det["name"]
                node = getattr(
                    get_model_element(model, det["component"]).mech, det["motion"]
                )
                el = detectors.MotionDetector(name, node)
                el._legacy_script_line_number = det["lineno"]
                model.add(el)

        # Apply attributes
        for block, d in blocks.items():
            for k, v in d["attributes"].items():
                component = get_model_element(model, k)
                for attr, val in v:
                    if attr == "mass":
                        name = component.name + "_free_mass"
                        model.add(
                            components.mechanical.FreeMass(name, component.mech, val)
                        )
                    else:
                        setattr(component, attr, val)

        cavities = []
        # Create cavities
        for block, d in blocks.items():
            for cav in d["cavities"]:
                try:
                    comp1 = nodes[cav["node1"]]
                except KeyError:
                    raise KeyError(f"Node '{cav['node11']}' not in model.")
                try:
                    comp2 = nodes[cav["node2"]]
                except KeyError:
                    raise KeyError(f"Node '{cav['node2']}' not in model.")
                node1 = getattr(get_model_element(model, comp1[0]), f"p{comp1[1]}")
                node2 = getattr(get_model_element(model, comp2[0]), f"p{comp2[1]}")

                if comp1[0] != comp2[0]:
                    el = components.Cavity(cav["name"], node1.o, node2.i)
                else:
                    el = components.Cavity(cav["name"], node1.o)

                el._legacy_script_line_number = cav["lineno"]
                model.add(el)
                cavities.append(el)

        gausses = []
        # Gauss commands
        for block, d in blocks.items():
            for gauss in d["gauss"]:
                try:
                    comp = nodes[gauss["node"]]
                except KeyError:
                    raise KeyError(f"Node '{gauss['node']}' not in model.")
                node = getattr(get_model_element(model, comp[0]), f"p{comp[1]}").o
                if gauss["component"] != comp[0]:
                    el = get_model_element(model, gauss["component"])
                    if isinstance(el, components.Space):
                        if node.port.space != el:
                            raise KeyError(
                                f"Invalid node '{gauss['node']}' for component '{gauss['component']}'."
                            )
                        # If a space was specified, flip the direction around
                        node = node.port.i
                    else:
                        raise KeyError(
                            f"Invalid node '{gauss['node']}' for component '{gauss['component']}'."
                        )

                gauss_kwargs = {}
                if "qy_re" in gauss:
                    gauss_kwargs["qx"] = gauss["qx_re"] + 1j * gauss["qx_im"]
                    gauss_kwargs["qy"] = gauss["qy_re"] + 1j * gauss["qy_im"]
                elif "qx_re" in gauss:
                    gauss_kwargs["q"] = gauss["qx_re"] + 1j * gauss["qx_im"]
                elif "w0y" in gauss:
                    gauss_kwargs["w0x"] = gauss["w0x"]
                    gauss_kwargs["zx"] = gauss["zx"]
                    gauss_kwargs["w0y"] = gauss["w0y"]
                    gauss_kwargs["zy"] = gauss["zy"]
                else:
                    gauss_kwargs["w0"] = gauss["w0x"]
                    gauss_kwargs["z"] = gauss["zx"]

                gauss_obj = components.Gauss(gauss["name"], node, **gauss_kwargs)
                model.add(gauss_obj)
                gausses.append(gauss_obj)

        # Want to keep Finesse 2 behaviour as much as possible in legacy parsing, so
        # here we set the priority values of the dependencies based on the order in
        # which the cavities and gausses were parsed to be consistent with Finesse 2
        Ndeps = len(cavities) + len(gausses)
        for i, dep in enumerate(cavities + gausses):
            dep.priority = Ndeps - i

        # Max TEM
        for block, d in blocks.items():
            maxtem = parse_parameter(d["maxtem"])
            if maxtem == "off" or maxtem is None:
                model.switch_off_homs()
            else:
                model.select_modes(maxtem=maxtem)

        # Phase command
        for block, d in blocks.items():
            model.phase_level = parse_parameter(d["phase"])

        # Lambda command
        for block, d in blocks.items():
            if d["lambda"] is not None:
                model.lambda0 = d["lambda"]

        # Retrace command
        for block, d in blocks.items():
            if d["retrace"] is None or d["retrace"] == "":
                continue
            elif d["retrace"] == "off":
                model.retrace = False
            else:
                LOGGER.warning(f"Unknown retrace argument '{d['retrace']}', ignoring")

        # Startnode command
        for block, d in blocks.items():
            if d["startnode"] is not None:
                comp = nodes[d["startnode"]]
                node = getattr(get_model_element(model, comp[0]), f"p{comp[1]}").o
                associated_gauss = model.gauss_commands.get(node)
                if associated_gauss is None:
                    LOGGER.error(
                        "startnode %s does not correspond to any Gauss command.",
                        node.full_name,
                    )

                # Get the current maximum priority (guaranteed to be first dependency)
                max_priority = model.trace_order[0].priority
                # Set the startnode Gauss as the new highest priority
                associated_gauss.priority = max_priority + 1

        # Input TEMs
        for block, d in blocks.items():
            for tem in d["tems"]:
                args = dict(tem)
                component = get_model_element(model, args.pop("component"))
                component.tem(**args)

        # Photo-detector types
        for block, d in blocks.items():
            for pdtype in d["pdtypes"]:
                args = dict(pdtype)
                detector = get_model_element(model, args.pop("detector"))

                if not isinstance(detector, detectors.PowerDetector):
                    raise ValueError("Cannot apply pdtype to a non pd detector.")

                name = detector.name
                node = detector.node
                # just getting the first letter of type now (either 'x' or 'y')
                # to create a SplitPD with x-split or y-split beats
                direction = args["type"][0]

                # get the demodulation frequencies and phases
                max_demodulations = 10
                freqs = []
                phases = []
                for i in range(max_demodulations):
                    f = getattr(detector, f"f{i}", None)
                    p = getattr(detector, f"phase{i}", None)

                    if f is not None and f.value is not None:
                        freqs.append(f)
                    if p is not None and p.value is not None:
                        phases.append(p)

                model.remove(detector)
                model.add(
                    detectors.SplitPD(name, node, direction, freqs=freqs, phases=phases)
                )

        # Detector masks
        for block, d in blocks.items():
            for mask in d["masks"]:
                args = dict(mask)
                detector = get_model_element(model, args.pop("detector"))
                detector.add_to_mask(**args)

        analyses = []

        # Xaxis
        camera_replacement = None
        for block, d in blocks.items():
            xaxis = d["xaxis"]
            x2axis = d["x2axis"]
            x3axis = d["x3axis"]
            if xaxis is None:
                continue

            xaxis["steps"] = int(xaxis["steps"])

            if xaxis["component"] in model.alternate_name_map:
                comp = model.alternate_name_map[xaxis["component"]]
            else:
                comp = get_model_element(model, f"{xaxis['component']}")

            # store this before converting with attribute getter below
            # as we want to use this for camera axis sweep checking
            xax_param = xaxis["parameter"]

            # phase1 & f1 for a pd1 have had the 1 removed, so rename them here
            if isinstance(comp, detectors.PowerDetectorDemod1) and xax_param[-1] == "1":
                xax_param = xax_param[:-1]

            xaxis["parameter"] = getattr(comp, xax_param)
            # cannot scan axes of Pixel, this is now done via ScanLine or Image type
            # cameras so we begin the process of transforming to one of these here
            if not (
                isinstance(comp, detectors.camera.Pixel) and xax_param in ("x", "y")
            ):
                model.alternate_name_map["x1"] = xaxis["parameter"].ref
                model.alternate_name_map["mx1"] = -xaxis["parameter"].ref

            del xaxis["component"]

            if xaxis["starred"]:
                xaxis["offset"] = xaxis["parameter"].value
            else:
                xaxis["offset"] = 0

            if isinstance(comp, detectors.camera.Pixel) and xax_param in ("x", "y"):
                LOGGER.info(
                    "Found an xaxis scan over the %s parameter of the "
                    "Pixel detector %s --> Replacing this detector and axis "
                    "sweep with a ScanLine detector.",
                    xax_param,
                    comp.name,
                )

                node = comp.node
                lim = [xaxis["min"], xaxis["max"]]
                npts = xaxis["steps"] + 1
                direction = xax_param

                if hasattr(comp, "f"):
                    f = comp.f.value
                else:
                    f = None

                # remove the Pixel detectors and add a ScanLine detector
                # in replacement of the axis sweep
                model.remove(comp)

                if f is None:
                    camera_replacement = detectors.CCDScanLine(
                        comp.name,
                        node,
                        npts,
                        xlim=lim if direction == "x" else None,
                        ylim=lim if direction == "y" else None,
                    )
                else:
                    camera_replacement = detectors.FieldScanLine(
                        comp.name,
                        node,
                        npts,
                        xlim=lim if direction == "x" else None,
                        ylim=lim if direction == "y" else None,
                        f=f,
                    )
                model.add(camera_replacement)

            if x2axis is None:
                if camera_replacement is None:
                    analyses.append(
                        analysis.actions.Xaxis(
                            xaxis["parameter"],
                            xaxis["scale"],
                            xaxis["min"],
                            xaxis["max"],
                            xaxis["steps"],
                            relative=xaxis["offset"],
                        )
                    )
            else:
                x2axis["steps"] = int(x2axis["steps"])

                if x2axis["component"] in model.alternate_name_map:
                    comp = model.alternate_name_map[x2axis["component"]]
                else:
                    comp = get_model_element(model, f"{x2axis['component']}")

                if isinstance(comp, detectors.camera.ScanLine):
                    if xax_param in ("x", "y") and x2axis["parameter"] in ("x", "y"):
                        LOGGER.info(
                            "Found an x2axis scan over the %s parameter of the previous "
                            "Pixel detector %s --> Replacing the ScanLine detector "
                            "added previously with a ComplexCamera.",
                            xax_param,
                            comp.name,
                        )

                        node = comp.node
                        x1ax_dir = xax_param
                        x2ax_dir = x2axis["parameter"]

                        if x1ax_dir == x2ax_dir:
                            raise ValueError(
                                "Cannot scan the same axis of the beam analyser twice."
                            )

                        if x1ax_dir == "x":
                            xlim = [xaxis["min"], xaxis["max"]]
                            ylim = [x2axis["min"], x2axis["max"]]
                        else:
                            xlim = [x2axis["min"], x2axis["max"]]
                            ylim = [xaxis["min"], xaxis["max"]]

                        npts = xaxis["steps"] + 1

                        if hasattr(comp, "f"):
                            f = comp.f.value
                        else:
                            f = None

                        # remove the ScanLine detector and add a ComplexCamera
                        # in replacement of axes sweeps
                        model.remove(comp)

                        if f is None:
                            camera_replacement = detectors.CCD(
                                comp.name, node, xlim, ylim, npts,
                            )
                        else:
                            camera_replacement = detectors.FieldCamera(
                                comp.name, node, xlim, ylim, npts, f
                            )
                        model.add(camera_replacement)

                else:
                    x2axis["parameter"] = getattr(comp, f"{x2axis['parameter']}")
                    model.alternate_name_map["x2"] = x2axis["parameter"].ref
                    model.alternate_name_map["mx2"] = -x2axis["parameter"].ref

                del x2axis["component"]

                if x2axis["starred"]:
                    x2axis["offset"] = x2axis["parameter"].value
                else:
                    x2axis["offset"] = 0
                if x3axis is None:
                    if camera_replacement is None:
                        analyses.append(
                            analysis.actions.X2axis(
                                xaxis["parameter"],
                                xaxis["scale"],
                                xaxis["min"],
                                xaxis["max"],
                                xaxis["steps"],
                                x2axis["parameter"],
                                x2axis["scale"],
                                x2axis["min"],
                                x2axis["max"],
                                x2axis["steps"],
                            )
                        )
                    elif isinstance(camera_replacement, detectors.camera.ScanLine):
                        analyses.append(
                            analysis.actions.Xaxis(
                                x2axis["parameter"],
                                x2axis["scale"],
                                x2axis["min"],
                                x2axis["max"],
                                x2axis["steps"],
                                relative=x2axis["offset"],
                            )
                        )
                else:
                    x3axis["steps"] = int(x3axis["steps"])

                    if x3axis["component"] in model.alternate_name_map:
                        comp = model.alternate_name_map[x3axis["component"]]
                    else:
                        comp = get_model_element(model, f"{x3axis['component']}")

                    x3axis["parameter"] = getattr(comp, f"{x3axis['parameter']}")
                    model.alternate_name_map["x3"] = x3axis["parameter"].ref
                    model.alternate_name_map["mx3"] = -x3axis["parameter"].ref

                    del x3axis["component"]

                    if x3axis["starred"]:
                        x3axis["offset"] = x3axis["parameter"].value
                    else:
                        x3axis["offset"] = 0

                    if camera_replacement is None:
                        analyses.append(
                            analysis.actions.X3axis(
                                xaxis["parameter"],
                                xaxis["scale"],
                                xaxis["min"],
                                xaxis["max"],
                                xaxis["steps"],
                                x2axis["parameter"],
                                x2axis["scale"],
                                x2axis["min"],
                                x2axis["max"],
                                x2axis["steps"],
                                x3axis["parameter"],
                                x3axis["scale"],
                                x3axis["min"],
                                x3axis["max"],
                                x3axis["steps"],
                            )
                        )
                    elif isinstance(camera_replacement, detectors.camera.ScanLine):
                        analyses.append(
                            analysis.actions.X2axis(
                                x2axis["parameter"],
                                x2axis["scale"],
                                x2axis["min"],
                                x2axis["max"],
                                x2axis["steps"],
                                x3axis["parameter"],
                                x3axis["scale"],
                                x3axis["min"],
                                x3axis["max"],
                                x3axis["steps"],
                            )
                        )
                    elif isinstance(camera_replacement, detectors.camera.Image):
                        analyses.append(
                            analysis.actions.Xaxis(
                                x3axis["parameter"],
                                x3axis["scale"],
                                x3axis["min"],
                                x3axis["max"],
                                x3axis["steps"],
                                relative=x3axis["offset"],
                            )
                        )

        # Yaxis
        for block, d in blocks.items():
            for yaxis in d["yaxis"]:
                if model.yaxis is not None:
                    raise ValueError("Cannot have more than one yaxis command.")
                model.yaxis = yaxis

        # Sets
        for block, d in blocks.items():
            for name, cmd in d["sets"].items():
                comp = get_model_element(model, cmd["component"])
                if isinstance(comp, detectors.Detector):
                    raise ValueError(
                        "Finesse 3 does not support using 'set' with the output of a detector. If "
                        "you are using this in combination with a 'lock' command, please switch "
                        "to the new syntax and use the 'lock' command as defined there."
                    )
                model.alternate_name_map[name] = getattr(
                    get_model_element(model, cmd["component"]), cmd["parameter"]
                )

        # Funcs
        for block, d in blocks.items():
            for name, func in d["functions"].items():
                model.alternate_name_map[name] = parse_parameter(func)

        # Puts
        scanning_im_ax = None
        for block, d in blocks.items():
            for put in d["puts"]:
                value = parse_parameter(put["variable"])
                if put["add"]:
                    component = get_model_element(model, put["component"])
                    param = getattr(component, put["parameter"])
                    setattr(component, put["parameter"], param + value)
                else:
                    component = get_model_element(model, put["component"])

                    if isinstance(component, components.SignalGenerator):
                        setattr(model.fsig, put["parameter"], value)
                    elif isinstance(component, detectors.camera.Pixel):
                        if isinstance(
                            camera_replacement, detectors.camera.ScanLine
                        ) and put["parameter"] in ("x", "y"):
                            model.remove(component)

                            if camera_replacement.direction == "x":
                                npts = camera_replacement.x.shape[0]
                            else:
                                npts = camera_replacement.y.shape[0]

                            if hasattr(component, "f"):
                                model.add(
                                    detectors.FieldScanLine(
                                        component.name,
                                        component.node,
                                        npts,
                                        camera_replacement.x,
                                        camera_replacement.y,
                                        camera_replacement.xlim,
                                        camera_replacement.ylim,
                                        component.f.value,
                                    )
                                )
                            else:
                                model.add(
                                    detectors.CCDScanLine(
                                        component.name,
                                        component.node,
                                        npts,
                                        camera_replacement.x,
                                        camera_replacement.y,
                                        camera_replacement.xlim,
                                        camera_replacement.ylim,
                                    )
                                )

                        elif isinstance(camera_replacement, detectors.camera.Image):
                            if scanning_im_ax is None:
                                scanning_im_ax = put["parameter"]
                                continue
                            elif (
                                scanning_im_ax == "x"
                                and put["parameter"] == "y"
                                or scanning_im_ax == "y"
                                and put["parameter"] == "x"
                            ):
                                model.remove(component)

                                if hasattr(component, "f"):
                                    model.add(
                                        detectors.FieldCamera(
                                            component.name,
                                            component.node,
                                            camera_replacement.x,
                                            camera_replacement.y,
                                            camera_replacement.x.shape[0],
                                            component.f.value,
                                        )
                                    )
                                else:
                                    model.add(
                                        detectors.CCD(
                                            component.name,
                                            component.node,
                                            camera_replacement.x,
                                            camera_replacement.y,
                                            camera_replacement.x.shape[0],
                                        )
                                    )
                            else:
                                setattr(component, put["parameter"], value)
                    else:
                        setattr(component, put["parameter"], value)

        scales = {}
        for block, d in blocks.items():
            for scale in d["scales"]:
                v = scale["value"]
                if isinstance(v, str):
                    import numpy as np

                    s = v.lower()
                    if s == "deg":
                        scale["value"] = 180 / np.pi
                    elif s == "rad":
                        scale["value"] = np.pi / 180
                    elif s == "meter":
                        scale["value"] = 2 * np.pi / model.lambda0
                    else:
                        LOGGER.error(
                            f"Scale type '{v}' not recognised, not performing scaling."
                        )
                        scale["value"] = 1
                comp = scale["component"]
                if comp is not None:
                    if comp in scales:
                        scales[comp] *= scale["value"]
                    else:
                        scales[comp] = scale["value"]
                else:
                    for det in model.detectors:
                        if det.name in scales:
                            scales[det.name] *= scale["value"]
                        else:
                            scales[det.name] = scale["value"]
        if len(scales) > 0:
            analyses[0] = analysis.actions.Serial(
                analyses[0], analysis.actions.Scale("Scale", scales)
            )

        if len(analyses) > 1:
            raise NotImplementedError(
                "Handling multiple analyses for legacy parsing not implemented yet"
            )

        if not analyses:
            analyses.append(analysis.actions.Noxaxis())

        model.analysis = analyses[0]

        def sort(item):
            has = hasattr(item[1], "_legacy_script_line_number")
            return (not has, item[1]._legacy_script_line_number if has else None)

        model._Model__elements = dict(sorted(model.elements.items(), key=sort))
        for el in model.elements.values():
            if hasattr(el, "_legacy_script_line_number"):
                del el._legacy_script_line_number
        return model


class _KatLEX(Lexer):
    """Kat file lexer, default state."""

    tokens = {
        "AMPLITUDE_DETECTOR",
        "ATTRIBUTE",
        "BEAM_DETECTOR",
        "BEAM_PROPERTY_DETECTOR",
        "BEAM_SPLITTER",
        "CAVITY",
        "COMMENT_START",
        "CONSTANT",
        "DIRECTIONAL_BEAM_SPLITTER",
        "SOURCE_FREQUENCY",
        "FTBLOCK_END",
        "FTBLOCK_START",
        "FUNCTION",
        "FSIG",
        "GAUSS",
        "GNUPLOT_START",
        "GOUY",
        "ISOLATOR",
        "LAMBDA",
        "LASER",
        "LENS",
        "LOCK",
        "MASK",
        "MAXTEM",
        "MIRROR",
        "MODULATOR",
        "MOTION_DETECTOR",
        "NOXAXIS",
        "PDTYPE",
        "PHASE",
        "POWER_DETECTOR",
        "PUT",
        "QUANTUM_NOISE_DETECTOR",
        "QUANTUM_SHOT_NOISE_DETECTOR",
        "RETRACE",
        "SCALE",
        "SET",
        "SPACE",
        "SQUEEZER",
        "STARTNODE",
        "TEM",
        "VARIABLE",
        "XAXIS",
        "X2AXIS",
        "X3AXIS",
        "YAXIS",
    }

    @_(
        "color",
        "conf",
        "debug",
        "frequency",
        "gnuterm",
        "noplot",
        "pause",
        "printmatrix",
        "pyterm",
        "showiterate",
        "time",
        "trace",
        "width",
    )
    def obsolete(self, t):
        self.warnings.append(
            (f"Command '{t.value}' is obsolete, ignoring.", self.lineno, 1)
        )
        line = self.text.split("\n")[self.lineno - 1]
        self.index += len(line) - len(t.value)

    @_(
        "multi",
        "pdS",
        "pdN",
        "hd",
        "qd",
        "sd",
        "qhd",
        "qhdS",
        "qhdN",
        "pgaind",
        "fadd",
        "map",
        "knm",
        "smotion",
        "vacuum",
        "tf",
        "tf2",
        "func",
        "diff",
        "deriv_h",
    )
    def not_implemented(self, t):
        self.warnings.append(
            (f"Command '{t.value}' not yet implemented.", self.lineno, 1)
        )
        line = self.text.split("\n")[self.lineno - 1]
        self.index += len(line) - len(t.value)

    # In order to allow components which are substrings of other components
    # (e.g. 'l' and 'lens'), these should be sorted alphabetically and then in
    # length order, such that no string is a substring of one that comes later
    # in the list.
    AMPLITUDE_DETECTOR = r"ad\s"
    ATTRIBUTE = r"attr\s"
    BEAM_SPLITTER = r"(beamsplitter|bs)[1-2]?\s"
    BEAM_DETECTOR = r"beam\s"
    BEAM_PROPERTY_DETECTOR = r"bp\s"
    CAVITY = r"cav(ity)?\s"
    COMMENT_START = r"/\*"
    CONSTANT = r"const\s"
    DIRECTIONAL_BEAM_SPLITTER = r"dbs\s"
    SOURCE_FREQUENCY = r"freq\s"
    FSIG = r"fsig\s"
    FTBLOCK_END = r"FTend"
    FTBLOCK_START = r"FTblock"
    FUNCTION = r"func\s"
    GAUSS = r"gauss\*?\s"
    GNUPLOT_START = r"(?i)GNUPLOT"
    GOUY = r"gouy\s"
    ISOLATOR = r"(isol|diode)\s"
    LAMBDA = r"lambda0?\s"
    LENS = r"lens\s"
    LOCK = r"lock\*?\s"
    LASER = r"(laser|light|l)\s"
    MASK = r"mask\s"
    MAXTEM = r"maxtem\s"
    MODULATOR = r"mod\s"
    MIRROR = r"(mirror|m[1-2]?)\s"
    NOXAXIS = r"noxaxis"
    PDTYPE = r"pdtype\s"
    PHASE = r"phase\s"
    POWER_DETECTOR = r"pd[0-9]*\s"
    PUT = r"put\*?\s"
    QUANTUM_NOISE_DETECTOR = r"qnoisedS?\s"
    QUANTUM_SHOT_NOISE_DETECTOR = r"qshotS?\s"
    RETRACE = r"retrace"
    STARTNODE = r"startnode\s"
    SCALE = r"scale\s"
    SET = r"set\s"
    SQUEEZER = r"sq\s"
    SPACE = r"s\s"
    TEM = r"tem\s"
    VARIABLE = r"variable\s"
    X3AXIS = r"x3axis\*?\s"
    X2AXIS = r"x2axis\*?\s"
    XAXIS = r"xaxis\*?\s"
    MOTION_DETECTOR = r"xd\s"
    YAXIS = r"yaxis\s"

    # Ignored patterns.
    ignore = "[ \t]"
    ignore_comment = "#.*"
    ignore_comment2 = "%((?!FT).)*"

    def __init__(self):
        self.reset()

    def reset(self):
        self.errors = []
        self.warnings = []

    @_(r"\n+")
    def ignore_newline(self, t):
        self.lineno += t.value.count("\n")

    def error(self, t):
        line = t.value.split("\n")[0]
        command = line.split(" ")[0]
        self.errors.append(
            (f"Command '{command}' unrecognised", self.lineno, self.index)
        )
        self.index += len(line)
        return t

    def eof(self, t):
        print("EOF")

    # Command type tokens.
    def AMPLITUDE_DETECTOR(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def ATTRIBUTE(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def BEAM_DETECTOR(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def BEAM_PROPERTY_DETECTOR(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def BEAM_SPLITTER(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def CAVITY(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def COMMENT_START(self, t):
        self.push_state(_KatCommentLEX)

    def CONSTANT(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def DIRECTIONAL_BEAM_SPLITTER(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def SOURCE_FREQUENCY(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def FTBLOCK_START(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def FTBLOCK_END(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def FUNCTION(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def FSIG(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def GAUSS(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def GNUPLOT_START(self, t):
        self.push_state(_KatGnuplotLEX)

    def GOUY(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def ISOLATOR(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def LAMBDA(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def LASER(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def LENS(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def LOCK(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def MASK(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def MAXTEM(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def MIRROR(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def MODULATOR(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def MOTION_DETECTOR(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def NOXAXIS(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def PDTYPE(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def PHASE(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def POWER_DETECTOR(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def PUT(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def QUANTUM_NOISE_DETECTOR(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def QUANTUM_SHOT_NOISE_DETECTOR(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def RETRACE(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def SCALE(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def SET(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def SPACE(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def SQUEEZER(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def STARTNODE(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def TEM(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def VARIABLE(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def XAXIS(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def X2AXIS(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def X3AXIS(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t

    def YAXIS(self, t):
        self.push_state(_KatComponentLEX)
        t.value = t.value.strip()
        return t


class _KatCommentLEX(Lexer):
    """Kat file lexer, comment state."""

    tokens = {"END"}

    @_(r"\n+")
    def ignore_newline(self, t):
        self.lineno += t.value.count("\n")

    def error(self, t):
        self.index += 1
        return

    @_(r"\*/")
    def END(self, t):
        self.pop_state()


class _KatGnuplotLEX(_KatCommentLEX):
    """Kat file lexer, gnuplot state."""

    tokens = {"END"}

    @_("END")
    def END(self, t):
        self.pop_state()


class _KatComponentLEX(Lexer):
    """Kat file lexer, component state."""

    tokens = {"FUNCTIONSTRING", "NUMBER", "STRING"}
    # Top level tokens.
    FUNCTIONSTRING = r"=[^=\n#]+"

    ignore = " \t"
    ignore_comment = r"\#.*"
    ignore_comment2 = "%.*"

    @_(r"\n+")
    def ignore_newline(self, t):
        self.lineno += t.value.count("\n")
        self.pop_state()

    @_(r"[$0-9\-+*(][a-zA-Z0-9_\-+*$().]+")
    def NUM_PARAM(self, t):
        if "$" not in t.value:
            # This is a number
            t.type = "NUMBER"
            return self.NUMBER(t)
        t.type = "NUMBER"
        return t

    # Number token including scientific notation, float,
    # or +/- inf (all states). Alternatively, any string starting with $
    @_(
        r"[+-]?inf",
        r"[+-]?(\d+\.\d*|\d*\.\d+|\d+)([eE]-?\d*\.?\d*)?([pnumkMGT])?",
        r"\$[.\s]+",
    )
    def NUMBER(self, t):
        if t.value.startswith("$"):
            return t
        if re.match(".*[pnumkMGT]$", t.value):
            t.value = t.value.replace("p", "e-12")
            t.value = t.value.replace("n", "e-9")
            t.value = t.value.replace("u", "e-6")
            t.value = t.value.replace("m", "e-3")
            t.value = t.value.replace("k", "e3")
            t.value = t.value.replace("M", "e6")
            t.value = t.value.replace("G", "e9")
            t.value = t.value.replace("T", "e12")

        # Check for numbers ending in "e", like "1e", which Python's float cannot handle.
        if t.value.endswith("e"):
            t.value += "0"

        if "j" in t.value:
            t.value = complex(t.value)
        else:
            t.value = float(t.value)
            if t.value.is_integer():
                t.value = int(t.value)
        return t

    @_(r"[a-zA-Z_][a-zA-Z0-9_:+-]*\*?", "inf")
    def STRING(self, t):
        if t.value == "inf":
            t.type = "NUMBER"
            return self.NUMBER(t)
        return t

    def error(self, t):
        line = t.value.split("\n")[0].split(" ")[0]
        self.errors.append(
            (f"Illegal character '{t.value[0]}'", self.lineno, self.index)
        )
        self.index += len(line)
        return t


class _KatYACC(Parser):
    """Kat file parser."""

    tokens = set.union(_KatLEX.tokens, _KatComponentLEX.tokens)
    tokens.remove("COMMENT_START")
    tokens.remove("GNUPLOT_START")

    # Setting STRING and NUMBER to have the same precedence and be right-associative solves the
    # shift/reduce conflict caused by the frequency_list rule in favour of shifting. We must then
    # be careful to extract the final string, as for e.g. the powerdetector, this is actually a
    # node name, not a phase.
    precedence = (("right", "STRING", "NUMBER"),)

    def __init__(self):
        self.reset()

    def reset(self):
        """Delete all parsed code, resetting the parser to a newly constructed state."""
        self.noxaxis = False
        self.block = None
        self.blocks = OrderedDict()
        self.blocks[self.block] = self._default_components()
        self.errors = []

    def _default_components(self):
        return {
            # Default simulation components.
            "lasers": [],
            "spaces": [],
            "mirrors": [],
            "beamsplitters": [],
            "directional_beamsplitters": [],
            "isolators": [],
            "modulators": [],
            "lenses": [],
            "amplitude_detectors": [],
            "beam_detectors": [],
            "beam_property_detectors": [],
            "gouy": [],
            "motion_detectors": [],
            "power_detectors": [],
            "quantum_noise_detectors": [],
            "cavities": [],
            "squeezers": [],
            # Non-component commands
            "frequencies": [],
            "constants": {},
            "attributes": {},
            "variables": {},
            "functions": {},
            "sets": {},
            "fsigs": [],
            "gauss": [],
            "locks": [],
            "puts": [],
            "pdtypes": [],
            "maxtem": None,
            "phase": 3,
            "retrace": None,
            "startnode": None,
            "scales": [],
            "tems": [],
            "masks": [],
            "xaxis": None,
            "x2axis": None,
            "x3axis": None,
            "yaxis": [],
            "lambda": None,
        }

    @_("instruction", "statement instruction")
    def statement(self, p):
        pass

    # List of one or more numbers
    @_("NUMBER", "number_list NUMBER")
    def number_list(self, p):
        if len(p) == 1:
            return [p.NUMBER]
        else:
            p.number_list.append(p.NUMBER)
            return p.number_list

    # List of one or more strings
    @_("STRING", "string_list STRING")
    def string_list(self, p):
        if len(p) == 1:
            return [p.STRING]
        else:
            p.string_list.append(p.STRING)
            return p.string_list

    # List of frequency-phase pairs
    @_(
        "NUMBER STRING",
        "NUMBER NUMBER",
        "frequency_list NUMBER STRING",
        "frequency_list NUMBER NUMBER",
    )
    def frequency_list(self, p):
        if len(p) == 2:
            return [[p[0], p[1]]]
        else:
            p.frequency_list.append([p[1], p[2]])
            return p.frequency_list

    # List of attribute-value pairs
    @_("STRING NUMBER", "attribute_list STRING NUMBER")
    def attribute_list(self, p):
        if len(p) == 2:
            return [[p.STRING, p.NUMBER]]
        else:
            p.attribute_list.append([p.STRING, p.NUMBER])
            return p.attribute_list

    # Frequencies can either be a number or a combination of source frequencies
    @_("NUMBER")
    def freq_num(self, p):
        return p[0]

    @_("FTBLOCK_START STRING")
    def instruction(self, p):
        if self.block is not None:
            LOGGER.warn(f"Already in FTblock {self.block}")
        if p.STRING in self.blocks:
            LOGGER.warn(f"Duplicate FTblock {p.STRING}")
        self.block = p.STRING
        self.blocks[self.block] = self._default_components()

    @_("FTBLOCK_END STRING")
    def instruction(self, p):
        if self.block != p.STRING:
            message = (
                f"Invalid command 'FTend {p.STRING}': currently in "
                f"FTblock '{self.block}'"
            )
            self.errors.append((message, p.lineno, p.index))
        self.block = None

    @_("SOURCE_FREQUENCY STRING NUMBER")
    def instruction(self, p):
        params = ["name", "f"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["frequencies"].append(dict(zip(params, values)))

    @_("LASER STRING NUMBER freq_num optnum STRING")
    def instruction(self, p):
        # Phase not specified.
        if p[4] is None:
            p[4] = 0

        params = ["name", "P", "f", "phase", "node"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["lasers"].append(dict(zip(params, values)))
        self.blocks[block]["lasers"][-1]["lineno"] = p.lineno

    @_("SQUEEZER STRING freq_num NUMBER NUMBER STRING")
    def instruction(self, p):
        params = ["name", "f", "db", "angle", "node"]
        values = [p[i] for i in range(1, len(p))]

        # Some minor rearranging required here to match the component constructor
        params.insert(2, params.pop(1))
        values.insert(2, values.pop(1))

        block = self.block
        self.blocks[block]["squeezers"].append(dict(zip(params, values)))
        self.blocks[block]["squeezers"][-1]["lineno"] = p.lineno

    @_("SPACE STRING NUMBER optnum STRING STRING")
    def instruction(self, p):
        params = ["name", "L", "nr", "node1", "node2"]
        values = [p[i] for i in range(1, len(p))]

        if values[2] is None:
            # Index of refraction not specified.
            # TODO:phil: should we be specifying 1 as the default here, or
            # checking for None in the space constructor later?
            values[2] = 1

        block = self.block
        self.blocks[block]["spaces"].append(dict(zip(params, values)))
        self.blocks[block]["spaces"][-1]["lineno"] = p.lineno

    @_("MIRROR STRING NUMBER NUMBER NUMBER STRING STRING")
    def instruction(self, p):
        params = ["name", "R", "T", "L", "phi", "node1", "node2"]
        values = [p[i] for i in range(1, len(p))]

        if p[0] == "m":
            # R / T
            values.insert(3, None)
        elif p[0] == "m1":
            # T / Loss.
            values.insert(1, None)
        elif p[0] == "m2":
            # R / Loss.
            values.insert(2, None)

        block = self.block
        self.blocks[block]["mirrors"].append(dict(zip(params, values)))
        self.blocks[block]["mirrors"][-1]["lineno"] = p.lineno

    @_("BEAM_SPLITTER STRING NUMBER NUMBER NUMBER NUMBER STRING STRING STRING STRING")
    def instruction(self, p):
        params = [
            "name",
            "R",
            "T",
            "L",
            "phi",
            "alpha",
            "node1",
            "node2",
            "node3",
            "node4",
        ]
        values = [p[i] for i in range(1, len(p))]

        if p[0].endswith("2"):
            # R / Loss.
            values.insert(2, None)
        elif p[0].endswith("1"):
            # T / Loss.
            values.insert(1, None)
        else:
            # R / T
            values.insert(3, None)

        block = self.block
        self.blocks[block]["beamsplitters"].append(dict(zip(params, values)))
        self.blocks[block]["beamsplitters"][-1]["lineno"] = p.lineno

    @_("DIRECTIONAL_BEAM_SPLITTER STRING STRING STRING STRING STRING")
    def instruction(self, p):
        params = ["name", "node1", "node2", "node3", "node4"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["directional_beamsplitters"].append(
            dict(zip(params, values))
        )
        self.blocks[block]["directional_beamsplitters"][-1]["lineno"] = p.lineno

    @_("ISOLATOR STRING NUMBER STRING STRING")
    def instruction(self, p):
        params = ["name", "S", "node1", "node2"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["isolators"].append(dict(zip(params, values)))
        self.blocks[block]["isolators"][-1]["lineno"] = p.lineno

    @_(
        "MODULATOR STRING freq_num NUMBER NUMBER STRING optnum STRING STRING",
        "MODULATOR STRING freq_num NUMBER STRING STRING optnum STRING STRING",
    )
    def instruction(self, p):
        params = ["name", "f", "midx", "order", "type", "phase", "node1", "node2"]
        values = [p[i] for i in range(1, len(p))]

        # TODO:phil: is no phase actually the same as 0 phase?
        if values[5] is None:
            values[5] = 0

        if values[3] == "s":
            values[3] = 1
            params.append("positive_only")
            values.append(True)

        block = self.block
        self.blocks[block]["modulators"].append(dict(zip(params, values)))
        self.blocks[block]["modulators"][-1]["lineno"] = p.lineno

    @_("LENS STRING NUMBER STRING STRING")
    def instruction(self, p):
        params = ["name", "f", "node1", "node2"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["lenses"].append(dict(zip(params, values)))
        self.blocks[block]["lenses"][-1]["lineno"] = p.lineno

    @_(
        "AMPLITUDE_DETECTOR STRING NUMBER NUMBER freq_num STRING",
        "AMPLITUDE_DETECTOR STRING freq_num STRING",
    )
    def instruction(self, p):
        params = ["name", "n", "m", "f", "node"]
        values = [p[i] for i in range(1, len(p))]

        if len(values) == 3:
            # Mode numbers not specified.
            values.insert(1, None)
            values.insert(2, None)

        block = self.block
        self.blocks[block]["amplitude_detectors"].append(dict(zip(params, values)))
        self.blocks[block]["amplitude_detectors"][-1]["lineno"] = p.lineno

    @_("BEAM_DETECTOR STRING optnum STRING")
    def instruction(self, p):
        params = ["name", "f", "node"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["beam_detectors"].append(dict(zip(params, values)))
        self.blocks[block]["beam_detectors"][-1]["lineno"] = p.lineno

    @_("BEAM_PROPERTY_DETECTOR STRING STRING STRING STRING")
    def instruction(self, p):
        params = ["name", "direction", "prop", "node"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["beam_property_detectors"].append(dict(zip(params, values)))
        self.blocks[block]["beam_property_detectors"][-1]["lineno"] = p.lineno

    @_("GOUY STRING STRING string_list")
    def instruction(self, p):
        params = ["name", "direction", "space_list"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["gouy"].append(dict(zip(params, values)))
        self.blocks[block]["gouy"][-1]["lineno"] = p.lineno

    @_("MOTION_DETECTOR STRING STRING STRING")
    def instruction(self, p):
        params = ["name", "component", "motion"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["motion_detectors"].append(dict(zip(params, values)))
        self.blocks[block]["motion_detectors"][-1]["lineno"] = p.lineno

    @_(
        "POWER_DETECTOR STRING STRING",
        "POWER_DETECTOR STRING frequency_list optstr",
        "POWER_DETECTOR STRING number_list STRING",
    )
    def instruction(self, p):
        # Parameters shared by all photodetectors.
        params = ["name", "node"]
        values = [p[i] for i in range(1, len(p))]

        ps = ["f{}", "phase{}"]
        if len(values) == 3:
            try:
                p.number_list
                for n in range(len(values[1])):
                    params.insert(n + 1, ps[n % 2].format(n // 2 + 1))
                values = [values[0], *values[1], values[2]]
            except AttributeError:
                if p.optstr is None:
                    # We've grabbed the node as part of the frequency list, so put it back in the
                    # right place
                    values[2] = values[1][-1].pop()
                for n in range(len(values[1])):
                    for m in range(len(values[1][n])):
                        params.insert(2 * n + m + 1, ps[m].format(n + 1))
                values = [
                    values[0],
                    *[param for pair in values[1] for param in pair],
                    values[2],
                ]

        block = self.block
        self.blocks[block]["power_detectors"].append(dict(zip(params, values)))
        self.blocks[block]["power_detectors"][-1]["lineno"] = p.lineno

    @_("QUANTUM_NOISE_DETECTOR STRING number_list STRING",)
    def instruction(self, p):
        # Parameters shared by all quantum noise detectors.
        params = ["name", "node"]
        values = [p[i] for i in range(1, len(p))]

        ps = ["f{}", "phase{}"]
        # Strip last demodulation, as this should always be at the signal frequency
        nf = len(values[1]) // 2 - 1
        for n in range(2 * nf):
            params.insert(n + 1, ps[n % 2].format(n // 2 + 1))
        values = [values[0], *values[1][1 : 2 * nf + 1], values[2]]

        params.insert(-1, "shot_only")
        values.insert(-1, False)

        params.insert(-1, "nsr")
        values.insert(-1, p[0].endswith("S"))

        block = self.block
        self.blocks[block]["quantum_noise_detectors"].append(dict(zip(params, values)))
        self.blocks[block]["quantum_noise_detectors"][-1]["lineno"] = p.lineno

    @_("QUANTUM_SHOT_NOISE_DETECTOR STRING number_list STRING",)
    def instruction(self, p):
        # Parameters shared by all quantum noise detectors.
        params = ["name", "node"]
        values = [p[i] for i in range(1, len(p))]

        ps = ["f{}", "phase{}"]
        # Strip last demodulation, as this should always be at the signal frequency
        nf = len(values[1]) // 2 - 1
        for n in range(2 * nf):
            params.insert(n + 1, ps[n % 2].format(n // 2 + 1))
        values = [values[0], *values[1][1 : 2 * nf + 1], values[2]]

        params.insert(-1, "shot_only")
        values.insert(-1, True)

        params.insert(-1, "nsr")
        values.insert(-1, p[0].endswith("S"))

        block = self.block
        self.blocks[block]["quantum_noise_detectors"].append(dict(zip(params, values)))
        self.blocks[block]["quantum_noise_detectors"][-1]["lineno"] = p.lineno

    @_("CAVITY STRING STRING STRING STRING STRING")
    def instruction(self, p):
        params = ["name", "component1", "node1", "component2", "node2"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["cavities"].append(dict(zip(params, values)))
        self.blocks[block]["cavities"][-1]["lineno"] = p.lineno

    @_("CONSTANT STRING NUMBER", "CONSTANT STRING STRING")
    def instruction(self, p):
        name = p[1]
        val = p[2]

        block = self.block
        self.blocks[block]["constants"]["$" + name] = val

    @_("ATTRIBUTE STRING attribute_list")
    def instruction(self, p):
        comp = p[1]
        attrs = p[2]

        block = self.block
        if comp in self.blocks[block]["attributes"]:
            self.blocks[block]["attributes"][comp].extend(attrs)
        else:
            self.blocks[block]["attributes"][comp] = attrs

    @_("VARIABLE STRING NUMBER", "VARIABLE STRING STRING")
    def instruction(self, p):
        name = p[1]
        value = p[2]

        block = self.block
        self.blocks[block]["variables"][name] = value

    @_("FUNCTION STRING FUNCTIONSTRING")
    def instruction(self, p):
        name = p[1]
        function_string = p[2]

        # Trim the starting "=" from the function string, and any whitespace
        block = self.block
        self.blocks[block]["functions"][name] = function_string[1:].strip()

    @_("FSIG STRING STRING optstr freq_num NUMBER optnum", "FSIG STRING NUMBER")
    def instruction(self, p):
        if len(p) == 3:
            params = ["name", "f"]
            values = [p[i] for i in range(1, len(p))]

            block = self.block
            self.blocks[block]["fsigs"].append(dict(zip(params, values)))
        else:
            if p.optnum is None:
                # Set default amplitude to 1
                p[-1] = 1
            params = ["name", "component", "mod_type", "f", "phase", "amp"]
            values = [p[i] for i in range(1, len(p))]

            block = self.block
            self.blocks[block]["fsigs"].append(dict(zip(params, values)))

    @_("LOCK STRING NUMBER NUMBER NUMBER")
    def instruction(self, p):
        params = ["name", "variable", "gain", "accuracy", "starred"]
        values = [p[i] for i in range(1, len(p))]

        if p[0].endswith("*"):
            values.append(True)
        else:
            values.append(False)
        block = self.block
        self.blocks[block]["locks"].append(dict(zip(params, values)))

    @_(
        "GAUSS STRING STRING STRING NUMBER NUMBER",
        "GAUSS STRING STRING STRING NUMBER NUMBER NUMBER NUMBER",
    )
    def instruction(self, p):
        values = [p[i] for i in range(1, len(p))]

        if p[0].endswith("*"):
            params = ["name", "component", "node", "qx_re", "qx_im", "qy_re", "qy_im"]
        else:
            params = ["name", "component", "node", "w0x", "zx", "w0y", "zy"]

        block = self.block
        self.blocks[block]["gauss"].append(dict(zip(params, values)))

    @_("PUT STRING STRING NUMBER")
    def instruction(self, p):
        params = ["component", "parameter", "variable", "add"]
        values = [p[i] for i in range(1, len(p))]

        if p[0].endswith("*"):
            values.append(True)
        else:
            values.append(False)
        block = self.block
        self.blocks[block]["puts"].append(dict(zip(params, values)))

    @_("SCALE NUMBER optstr", "SCALE STRING optstr")
    def instruction(self, p):
        params = ["value", "component"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["scales"].append(dict(zip(params, values)))

    @_("SET STRING STRING STRING")
    def instruction(self, p):
        name = p[1]
        params = ["component", "parameter"]
        values = [p[i] for i in range(2, len(p))]

        block = self.block
        self.blocks[block]["sets"][name] = dict(zip(params, values))

    @_("PDTYPE STRING STRING")
    def instruction(self, p):
        params = ["detector", "type"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["pdtypes"].append(dict(zip(params, values)))

    @_("MAXTEM NUMBER", "MAXTEM STRING")
    def instruction(self, p):
        block = self.block
        self.blocks[block]["maxtem"] = p[1]

    @_("PHASE NUMBER")
    def instruction(self, p):
        block = self.block
        self.blocks[block]["phase"] = p[1]

    @_("RETRACE optstr")
    def instruction(self, p):
        block = self.block
        self.blocks[block]["retrace"] = p[1] or ""

    @_("STARTNODE STRING")
    def instruction(self, p):
        block = self.block
        self.blocks[block]["startnode"] = p[1]

    @_("TEM STRING NUMBER NUMBER NUMBER NUMBER")
    def instruction(self, p):
        params = ["component", "n", "m", "factor", "phase"]
        values = [p[i] for i in range(1, len(p))]

        block = self.block
        self.blocks[block]["tems"].append(dict(zip(params, values)))

    @_("MASK STRING NUMBER NUMBER NUMBER")
    def instruction(self, p):
        params = ["detector", "modes", "factor"]
        values = [p[1], (p[2], p[3]), p[4]]

        block = self.block
        # ignoring factor as masks in Finesse 3 just zero the given field
        self.blocks[block]["masks"].append(dict(zip(params[:-1], values[:-1])))

    @_("LAMBDA NUMBER")
    def instruction(self, p):
        block = self.block
        self.blocks[block]["lambda"] = p.NUMBER

    @_("NOXAXIS")
    def instruction(self, p):
        self.noxaxis = True

    @_("XAXIS STRING STRING STRING NUMBER NUMBER NUMBER")
    def instruction(self, p):
        params = ["component", "parameter", "scale", "min", "max", "steps", "starred"]
        values = [p[i] for i in range(1, len(p))]

        if p[0].endswith("*"):
            values.append(True)
        else:
            values.append(False)

        block = self.block
        self.blocks[block]["xaxis"] = dict(zip(params, values))

    @_("X2AXIS STRING STRING STRING NUMBER NUMBER NUMBER")
    def instruction(self, p):
        params = ["component", "parameter", "scale", "min", "max", "steps", "starred"]
        values = [p[i] for i in range(1, len(p))]

        if p[0].endswith("*"):
            values.append(True)
        else:
            values.append(False)

        block = self.block
        self.blocks[block]["x2axis"] = dict(zip(params, values))

    @_("X3AXIS STRING STRING STRING NUMBER NUMBER NUMBER")
    def instruction(self, p):
        params = ["component", "parameter", "scale", "min", "max", "steps", "starred"]
        values = [p[i] for i in range(1, len(p))]

        if p[0].endswith("*"):
            values.append(True)
        else:
            values.append(False)

        block = self.block
        self.blocks[block]["x3axis"] = dict(zip(params, values))

    # TODO: placing the optstr before the STRING like in Finesse 2 causes a
    # shift/reduce conflict - can we solve this?
    @_("YAXIS STRING optstr")
    def instruction(self, p):
        params = ["scale", "axes"]
        values = [p[i] for i in range(1, len(p))]

        if values[1] is None:
            values.insert(0, "lin")

        block = self.block
        self.blocks[block]["yaxis"].append(dict(zip(params, values)))

    def error(self, p):
        if p is None:
            msg = "Unexpected end of file"
            self.errors.append((msg, None, None))
            return
        elif p.type == "ERROR":
            return
        msg = f"got unexpected token {p.value} of type {p.type}"
        self.errors.append((msg, p.lineno, p.index))

    @_("")
    def empty(self, p):
        pass

    @_("STRING")
    def optstr(self, p):
        return p.STRING

    @_("empty")
    def optstr(self, p):
        pass

    @_("NUMBER")
    def optnum(self, p):
        return p.NUMBER

    @_("empty")
    def optnum(self, p):
        pass


class KatParserError(ValueError):  # __NODOC__
    """Kat file parser error."""

    def __init__(self, errors, text, **kwargs):
        message = "\n"
        for error in errors:
            lineno = error[1]
            idx = error[2]
            if lineno is None:
                # There is no lineno, as this was an end-of-file error,
                # so assume error was on last non-empty lin
                line = re.findall(r"[^\s]", text)[-1]
                pos = len(text) - 1
            else:
                line = text.split("\n")[lineno - 1]
                pos = find_column(text, idx)
            expected = ""
            for pattern, exp in self.expected.items():
                if exp is not None and re.match(pattern, line) is not None:
                    expected = f", expected '{exp}'"
                    break
            message += f"{lineno}:{pos}: " + error[0] + expected + "\n"
            message += line + "\n"
            message += " " * (pos - 1) + "^\n"

        super().__init__(message.rstrip("\n"), **kwargs)

    expected = {
        r"ad\s": "ad name [n m] f node[*]",
        r"attr\s": "attr component parameter value",
        r"beam\s": "beam name [f] node[*]",
        r"bp\s": "bp name x/y parameter node",
        r"bs2\s": "bs2 name R L phi alpha node1 node2 node3 node4",
        r"bs1\s": "bs1 name T L phi alpha node1 node2 node3 node4",
        r"bs\s": "bs name R T phi alpha node1 node2 node3 node4",
        r"cav\s": "cav name component1 node component2 node",
        r"const\s": "const name value",
        r"dbs\s": "dbs name node1 node2 node3 node4",
        r"freq\s": None,
        r"FTblock": "FTblock name",
        r"FTend": "FTend name",
        r"func\s": "func name = function-string",
        r"fsig\s": ("fsig name component [type] f phase [amp]", "fsig name f"),
        r"gauss*\s": "gauss* name component node q [qy]",
        r"gauss\s": "gauss name component node w0 z [wy0 zy]",
        r"gouy\s": "gouy name x/y space-list",
        r"isol\s": "isol name S [Loss] node1 node2 [node3]",
        r"lambda\s": "lambda wavelength",
        r"lock\s": "lock name function/set gain accuracy [offset]",
        r"lens\s": "lens name f node1 node2",
        r"l\s": "l name I f [phase] node",
        r"maxtem\s": "maxtem order",
        r"mod\s": "mod name f midx order am/pm/yaw/pitch node1 node2",
        r"m2\s": "m2 name R L phi node1 node2",
        r"m1\s": "m1 name T L phi node1 node2",
        r"m\s": "m name R T phi node1 node2",
        r"noxaxis": None,
        r"pd\s": "pd[n] name [f1 [phase1 [f2 [phase2 [...] ] ] ] ] node[*]",
        r"phase\s": "phase 0-7",
        r"put\s": "put component parameter function/set/axis",
        r"qnoised\s": "qnoised name num_demods f1 phase1 [f2 phase2 [...]] node[*]",
        r"qshot\s": "qshot name num_demods f1 phase1 [f2 phase2 [...]] node[*]",
        r"retrace": "retrace [off|force]",
        r"s\s": "s name L [n] node1 node2",
        r"tem\s": "tem input n m factor phase",
        r"variable\s": "variable name value",
        r"xaxis\s": "xaxis component parameter lin/log min max steps",
        r"x2axis\s": "x2axis component parameter lin/log min max steps",
        r"yaxis\s": "yaxis [lin/log] abs:deg/db:deg/re:im/abs/db/deg",
    }


def find_column(text, index):  # __NODOC__
    last_cr = text.rfind("\n", 0, index)
    if last_cr < 0:
        last_cr = 0
    column = index - last_cr
    return column



"""Modal computation module

This module contains highly optimised code for each component for conducting modal simulations.
Component workspaces should register these fill and calculation methods as required.
"""



from finesse.components.node import NodeType, NodeDirection, Node
from finesse.components.general import Connector, ModelElement, DOFDefinition
from finesse.components.workspace import ConnectorWorkspace, Connections, ConnectionSetting
import numpy as np
from finesse.parameter import float_parameter


class DOFWorkspace(ConnectorWorkspace):
    def __init__(self, owner, sim, refill):
        super().__init__(owner, sim, refill, refill, Connections(), Connections())
        self.drives = None
        self.amplitudes = None


@float_parameter("DC", "DC state of degree of freedom")
class DegreeOfFreedom(Connector):
    def __init__(self, name, node, amplitude=1, *node_amp_pairs, DC=0):
        Connector.__init__(self, name)
        self._add_to_model_namespace = True
        self.__nodes = tuple((node, *node_amp_pairs[::2]))
        self.__amplitudes = tuple((amplitude, *node_amp_pairs[1::2]))
        self.DC = DC

        if len(self.drives) != len(self.amplitudes):
            raise Exception(
                f"Nodes and amplitudes were not the same length, {len(self.drives)} vs {len(self.amplitudes)}"
            )

        AC_type = None
        for node in self.drives:
            if not isinstance(node, DOFDefinition):
                raise Exception(
                    f"Degree of freedom ({name}) input `{node}` should be a {DOFDefinition}"
                )
            if not (node.AC.type == NodeType.ELECTRICAL or node.AC.type == NodeType.MECHANICAL):
                raise Exception(
                    f"Degree of freedom ({name}) input `{node}` should be an electrical or mechanical node"
                )
            if AC_type and AC_type != node.AC.type:
                raise Exception(f"Degree of freedom ({name}) input `{node}` should be the same type as other nodes, {AC_type}")
            else:
                AC_type = node.AC.type

        for amp in self.amplitudes:
            if not (np.isscalar(amp) and np.real(amp)):
                raise Exception(
                    f"Degree of freedom ({name}) amplitude `{amp}` is not a real number"
                )

        if AC_type:
            # Only add an AC port if there are some AC drives
            self._add_port("AC", NodeType.ELECTRICAL)
            self.AC._add_node("i", NodeDirection.INPUT)
            self._add_port("out", AC_type)

            for i, node in enumerate(self.drives):
                self.out._add_node(f"o{i}", None, node=self.drives[i].AC)
                self._register_node_coupling(f"AC_out{i}", self.AC.i, self.drives[i].AC)

    def _on_add(self, model):
        # Setup this DOf to set itself as an external
        # setter for the DC parameters it injects into
        model._on_pre_build.append(self._pre_build)
        model._on_unbuild.append(self._on_unbuild)

    def _pre_build(self):
        # Set up the DC parameters to be controlled externally, by this DOF element
        for node, amp in zip(self.drives, self.amplitudes):
            dc_param = node.DC
            if dc_param is not None:
                # Here we set the DC parameter associated with a node to track the
                # value of the DC parameter of this DOF.
                # mark that this element will be controlling the value of this parameter
                node.DC.set_external_setter(self, amp * self.DC.ref)

    def _on_unbuild(self):
        # need to remove our 
        for node, amp in zip(self.drives, self.amplitudes):
            dc_param = node.DC
            if dc_param is not None:
                node.DC.remove_external_setter(self, amp * self.DC.ref)

    @property
    def node(self):
        ":getter: Returns the node to be used for driving."
        return self.p1.i

    @property
    def drives(self):
        ":getter: Returns The nodes this degree of freedom drives."
        return tuple(self.__nodes)

    @property
    def amplitudes(self):
        ":getter: Returns the node amplitudes which a node is driven."
        return tuple(self.__amplitudes)

    @property
    def dc_enabled(self):
        """:getter: Returns True if all driving nodes have an associated DC parameter that can be varied."""
        return all((_.dc_parameter is not None for _ in self.drives))

    def _get_workspace(self, sim):
        if sim.signal:
            ws = DOFWorkspace(self, sim, False)
            ws.signal.add_fill_function(self.__fill, False)
            ws.drives = self.drives
            ws.amplitudes = np.array(self.amplitudes)
            for i, drive in enumerate(self.drives):
                if drive.AC.type == NodeType.MECHANICAL:
                    ws.amplitudes[i] /= sim.model_data.x_scale
            return ws
        else:
            return None

    def __fill(self, ws):
        for idx in range(len(ws.drives)):
            # Need to loop and determine if our connections have
            # been allocated or not
            mat_views = getattr(ws.signal.connections, "AC_out"+str(idx))
            if mat_views:
                # All connections are just their amplitude value
                # assumes no HOM couplings or anything between elec
                # and mechanical nodes
                mat_views[0][:] = ws.amplitudes[idx]


"""Test fixtures shared by all tests."""

import pytest


def pytest_addoption(parser):
    # Add an option to change the Hypothesis max_examples setting.
    parser.addoption(
        "--hypothesis-max-examples",
        action="store",
        default=None,
        help="set the Hypothesis max_examples setting",
    )


def pytest_configure(config):
    # Set Hypothesis max_examples.
    hypothesis_max_examples = config.getoption("--hypothesis-max-examples")
    if hypothesis_max_examples is not None:
        import hypothesis

        hypothesis.settings.register_profile(
            "hypothesis-overridden", max_examples=int(hypothesis_max_examples)
        )

        hypothesis.settings.load_profile("hypothesis-overridden")


@pytest.fixture(autouse=True)
def test_setup_and_teardown():
    """Fixture that runs before and after all tests."""
    from finesse.config import autoconfigure
    from finesse.datastore import invalidate

    # Reset to default configuration.
    autoconfigure()

    # Run the test.
    yield

    # Delete the global cache.
    invalidate()



"""Finesse command line interface."""

import click

# https://github.com/pallets/click/issues/430
from click_default_group import DefaultGroup

from . import __version__, PROGRAM, DESCRIPTION
from .cli import (
    run,
    info,
    syntax,
    config,
    convert,
    print_banner,
)


@click.group(cls=DefaultGroup, default="run", help=DESCRIPTION)
@click.option(
    "--banner",
    is_flag=True,
    default=False,
    callback=print_banner,
    expose_value=False,
    is_eager=True,
    help=print_banner.__doc__,
)
@click.version_option(version=__version__, prog_name=PROGRAM)
@click.pass_context
def cli(ctx):
    pass


cli.add_command(run)
cli.add_command(info)
cli.add_command(syntax)
cli.add_command(config)
cli.add_command(convert)



"""CLI config command tests."""

import pytest
from finesse.__main__ import config
from finesse.config import config_instance
from .util import sanitized_output


@pytest.fixture
def dumped_config(request, cli):
    """The dumped configuration table."""
    return sanitized_output(cli.invoke(config, ["--dump"]))


def test_paths(cli):
    """Test printing the config paths."""
    cli_result = cli.invoke(config)
    output = sanitized_output(cli_result)

    for path in config_instance().user_config_paths().values():
        assert str(path.resolve()) in output

    assert cli_result.exit_code == 0


@pytest.mark.parametrize(
    "key,value",
    [
        (k, v)
        for s in config_instance().sections()
        for k, v in config_instance().items(s)
    ],
)
def test_dump(dumped_config, key, value):
    """Test config dump."""
    assert key in dumped_config
    assert value in dumped_config


@pytest.mark.xfail(reason="The --yes noninteractive option is currently not supported")
def test_user_config_reset(monkeypatch, tmp_path, cli):
    """Test user config reset."""
    config = config_instance()

    # Mock the user config location and write the default user config there.
    monkeypatch.setattr(config, "user_config_dir", lambda: tmp_path)
    config.write_user_config()

    # String to write into the current (temporary) config to verify it gets reset.
    test_message = "# __TEST__"

    # Add test message to existing config and verify it's there.
    with config.user_config_path().open("w") as fobj:
        fobj.writelines([test_message])
    with config.user_config_path().open("r") as fobj:
        lines = fobj.readlines()
    assert test_message in lines

    # Reset the config and verify the test message is gone.
    cli_result = cli.invoke(config, ["--reset", "--yes"])
    assert cli_result.exit_code == 0
    with config.user_config_path().open("r") as fobj:
        lines = fobj.readlines()
    assert test_message not in lines



import re
import pytest
from finesse.script.exceptions import KatParsingError
from finesse.symbols import OPERATORS
from ....util import dedent_multiline


@pytest.fixture
def compiler(compiler, fake_element_cls):
    compiler.spec.register_element(
        "fake_element", {"setter": fake_element_cls, "getter": fake_element_cls}
    )
    # Have to use real Finesse operator here because the builder matches against Finesse operations.
    compiler.spec.binary_operators["-"] = OPERATORS["__sub__"]
    return compiler


@pytest.mark.parametrize(
    "script,element_def",
    (
        pytest.param(
            "fake_element myelement 1", ("myelement", 1), id="element-with-arg",
        ),
        pytest.param(
            "fake_element myelement a=1", ("myelement", 1), id="element-with-kwarg",
        ),
    ),
)
def test_element(model_matcher, fake_element_cls, script, element_def):
    model_matcher(script, [fake_element_cls(*element_def)])


@pytest.mark.parametrize(
    "script,element_defs",
    (
        pytest.param(
            dedent_multiline(
                """
                fake_element myelement1 1
                fake_element myelement2 1-&myelement1.a
                """
            ),
            [("myelement1", 1), ("myelement2", 0)],
        ),
    ),
)
def test_argument_reference(model_matcher, fake_element_cls, script, element_defs):
    model_matcher(script, [fake_element_cls(*defs) for defs in element_defs])


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            "fake_element myelement1 &myelement1.a",
            (
                "\nline 1: cannot set myelement1.a to self-referencing value myelement1.a\n"
                "-->1: fake_element myelement1 &myelement1.a\n"
                "                              ^^^^^^^^^^^^^"
            ),
            id="arg-self-ref",
        ),
        pytest.param(
            "fake_element myelement1 1-&myelement1.a",
            (
                "\nline 1: cannot set myelement1.a to self-referencing value (1-myelement1.a)\n"
                "-->1: fake_element myelement1 1-&myelement1.a\n"
                "                                ^^^^^^^^^^^^^"
            ),
            id="arg-expr-self-ref",
        ),
        pytest.param(
            "fake_element myelement1 a=&myelement1.a",
            (
                "\nline 1: cannot set myelement1.a to self-referencing value myelement1.a\n"
                "-->1: fake_element myelement1 a=&myelement1.a\n"
                "                                ^^^^^^^^^^^^^"
            ),
            id="kwarg-self-ref",
        ),
        pytest.param(
            "fake_element myelement1 a=1-&myelement1.a",
            (
                "\nline 1: cannot set myelement1.a to self-referencing value (1-myelement1.a)\n"
                "-->1: fake_element myelement1 a=1-&myelement1.a\n"
                "                                  ^^^^^^^^^^^^^"
            ),
            id="kwarg-expr-self-ref",
        ),
    ),
)
def test_directly_self_referencing_parameter_invalid(compiler, script, error):
    with pytest.raises(KatParsingError, match=re.escape(error)):
        compiler.compile(script)


@pytest.mark.parametrize(
    "script,element_values",
    (
        pytest.param(
            "fake_element myelement1 1 0.5-&myelement1.a",
            {"a": 1, "b": -0.5},
            id="arg-self-ref-to-arg",
        ),
        pytest.param(
            "fake_element myelement1 1 b=0.5-&myelement1.a",
            {"a": 1, "b": -0.5},
            id="kwarg-self-ref-to-arg",
        ),
        pytest.param(
            "fake_element myelement1 a=1 b=0.5-&myelement1.a",
            {"a": 1, "b": -0.5},
            id="kwarg-self-ref-to-kwarg",
        ),
        pytest.param(
            dedent_multiline(
                """
                fake_element myelement1 a=&myelement2.a b=0.5-&myelement1.a
                fake_element myelement2 a=0.5-&myelement2.b b=1
                """
            ),
            {"a": -0.5, "b": 1},
            id="second-order-self-ref",
        ),
    ),
)
def test_same_element_referencing_element(compiler, script, element_values):
    model = compiler.compile(script)
    for key, value in element_values.items():
        assert model.reduce_get_attr(f"myelement1.{key}").eval() == value


@pytest.mark.parametrize(
    "script,element_defs",
    (
        pytest.param(
            dedent_multiline(
                """
                fake_element myelement1 1
                fake_element myelement2 2
                fake_element myelement3 3
                """
            ),
            [("myelement1", 1), ("myelement2", 2), ("myelement3", 3),],
            id="elements",
        ),
    ),
)
def test_script(model_matcher, fake_element_cls, script, element_defs):
    model_matcher(script, [fake_element_cls(*defs) for defs in element_defs])



def test_link(model):
    """Test that the link command parses correctly."""
    model.parse(
        """
        laser l1
        mirror m1
        mirror m2
        link(l1, m1, m2)
        """
    )

    assert model.path(model.l1.p1.o, model.m2.p1.i)



# This test is based on
# https://finesse.docs.ligo.org/finesse3/getting_started/examples/4_geometrical_params.html
# and https://dcc.ligo.org/LIGO-T1900708
# This test checks that the geometric model
# and Finesse agree on predicting the spot
# position shift

import numpy as np
from numpy.testing import assert_allclose
import pytest
import finesse


def geometric(mirror, theta_low, theta_high, npoints, L=0.1, RcM2=0.5):

    g1 = 1  # g1 = 1 - L/RcM1 = 1 since RcM1 = infinity
    g2 = 1 - L / RcM2  # g2 for M2 mirror

    if mirror == 1:
        theta_1 = np.linspace(  # M1 tilt
            theta_low, theta_high, npoints + 1, endpoint=True
        )
        theta_2 = 0  # M2 tilt

    elif mirror == 2:
        theta_1 = 0  # M1 tilt
        theta_2 = np.linspace(  # M2 tilt
            theta_low, theta_high, npoints + 1, endpoint=True
        )
    else:
        raise NotImplementedError("Mirror should be 1 or 2 and an integer")

    # Resulant translation
    delta_x1 = (g2 / (1 - g1 * g2)) * L * theta_1 - (
        1 / (1 - g1 * g2)
    ) * L * theta_2  # Displacement on M1
    delta_x2 = (1 / (1 - g1 * g2)) * L * theta_1 - (
        g1 / (1 - g1 * g2)
    ) * L * theta_2  # Displacement on M2
    delta_theta = (delta_x2 - delta_x1) / L  # Angular displacement of optical axis

    print("g1g2 =", g1 * g2)  # Stability check
    print("Results for analytical solutions:")
    print("∆x1: ", max(delta_x1), min(delta_x1), " m")
    print("∆x2: ", max(delta_x2), min(delta_x2), " m")
    print(r"∆θ: ", max(delta_theta), min(delta_theta), " m")

    return delta_x1, delta_x2, delta_theta


def make_model(L=0.1, RcM2=0.5, T=1e-6):
    T = 1e-6

    # Build Finesse Model
    kat = finesse.Model()
    kat.parse(
        f"""
    laser laser P=1.0 f=0 phase=0

    # Mirrors
    mirror M1 R=(1-&M1.T) T={T} L=0.0 phi=0 Rc=[inf, inf] xbeta=0 ybeta=0
    mirror M2 R=(1-&M1.T) T=&M1.T L=0.0 phi=0 Rc=[{RcM2}, {RcM2}] xbeta=0 ybeta=0

    # Spaces
    space s0 portA=laser.p1 portB=M1.p1 L=1 nr=1 gouy_x=0 gouy_y=0
    space s_cav portA=M1.p2 portB=M2.p1 L=0.1 nr=1 gouy_x=0 gouy_y=0

    # Ampltitude Detectors
    amplitude_detector det00_1 node=M1.p2.i f=0 n=0 m=0
    amplitude_detector det00_2 node=M2.p1.o f=0 n=0 m=0
    amplitude_detector det10_1 node=M1.p2.i f=0 n=1 m=0
    amplitude_detector det10_2 node=M2.p1.o f=0 n=1 m=0

    # Beam Property Detectors
    beam_property_detector bp_n2 prop=0 node=M1.p2.i direction=x q_as_bp=false
    beam_property_detector bp_n3 prop=0 node=M2.p1.o direction=x q_as_bp=false
    beam_property_detector cavwaist prop=1 node=M1.p2.i direction=x q_as_bp=false

    # Photodiodes
    power_detector_dc pcirc node=M1.p2.o pdtype=none

    # Config
    cavity cavity1 source=M1.p2.o via=M2.p1.i priority=1
    lambda(1.064e-06)
    modes(maxtem=3)
    """
    )
    tsy = finesse.tracing.tools.propagate_beam(
        to_node=kat.M2.p1.o, from_node=kat.M1.p2.o, direction="y"
    )
    print(tsy.make_table(tablefmt="simple"))
    return kat


def run_simulation(mirror, theta_low, theta_high, npoints, kat):
    kat.parse(
        f"""
    # Actions
    xaxis(
        parameter=M{mirror:.0f}.xbeta,
        mode=lin,
        start={theta_low},
        stop={theta_high},
        steps={npoints},
        pre_step=none,
        post_step=none)
    """
    )
    return proccess_sim(kat)


def proccess_sim(kat):
    out = kat.run()
    # Comparison
    acirc = np.sqrt(out["pcirc"].real)  # circulating amplitude

    # intermodal phase HG10 -> HG00
    # at mirror 1
    intermodal_phase_1 = np.angle(out["det10_1"]) - np.angle(out["det00_1"])
    # at mirror 2
    intermodal_phase_2 = np.angle(out["det10_2"]) - np.angle(out["det00_2"])

    # Real part of \power normalised HG10 mode ampltiude
    # at M1
    a_shift_1 = (np.abs(out["det10_1"]) / acirc) * np.cos(intermodal_phase_1)
    # at M2
    a_shift_2 = (np.abs(out["det10_2"]) / acirc) * np.cos(intermodal_phase_2)

    # Imaginary part of power normalised HG10 mode ampltiude
    # at M1
    a_tilt_1 = (np.abs(out["det10_1"]) / acirc) * np.sin(intermodal_phase_1)
    # at M2
    a_tilt_2 = (np.abs(out["det10_2"]) / acirc) * np.sin(intermodal_phase_2)

    # Spot position shift
    # at M1
    dx_finesse_1 = out["bp_n2"].real * a_shift_1
    # at M2
    dx_finesse_2 = out["bp_n3"].real * a_shift_2

    # Wavefront tilt at
    # M1
    dtheta_1 = a_tilt_1 * kat.lambda0 / (np.pi * out["cavwaist"].real)
    # M2
    dtheta_2 = a_tilt_2 * kat.lambda0 / (np.pi * out["cavwaist"].real)

    print("Results for finesse solutions:")
    print(
        "Intermodal_1: ",
        max(intermodal_phase_1 / np.pi),
        min(dx_finesse_1 / np.pi),
        " pi",
    )
    print(
        "Intermodal_2: ",
        max(intermodal_phase_2 / np.pi),
        min(dx_finesse_2 / np.pi),
        " pi",
    )

    print("∆x1: ", max(dx_finesse_1), min(dx_finesse_1), " m")
    print("∆x2: ", max(dx_finesse_2), min(dx_finesse_2), " m")
    print(r"∆θ1: ", max(dtheta_1), min(dtheta_1), " m")
    print(r"∆θ2: ", max(dtheta_2), min(dtheta_2), " m")

    return (
        intermodal_phase_1,
        intermodal_phase_2,
        dx_finesse_1,
        dx_finesse_2,
        dtheta_1,
        dtheta_2,
        out.x[0],
    )


@pytest.mark.parametrize("T", [1e-6, 1e-7])
@pytest.mark.parametrize("L,RcM2", [(0.1, 0.5), (0.48, 0.5)])
@pytest.mark.parametrize("mirror", [1, 2])
def test(L, RcM2, T, mirror):

    # Mirror Tilt:
    npoints = 40
    theta_low = -1e-6
    theta_high = 1e-6

    delta_x1, delta_x2, delta_theta = geometric(mirror, theta_low, theta_high, npoints)

    kat = make_model(L=L, RcM2=RcM2, T=T)

    (
        intermodal_phase_1,
        intermodal_phase_2,
        dx_finesse_1,
        dx_finesse_2,
        dtheta_1,
        dtheta_2,
        xaxis,
    ) = run_simulation(mirror, theta_low, theta_high, npoints, kat)

    assert_allclose(delta_x1, dx_finesse_1, rtol=0, atol=1e-11)
    assert_allclose(delta_x2, dx_finesse_2, rtol=0, atol=1e-11)
    assert_allclose(delta_theta, dtheta_1, rtol=0, atol=1e-11)
    assert_allclose(delta_x1, dx_finesse_1, rtol=0, atol=1e-11)



from textwrap import dedent
from finesse.model import Model
from finesse.script.spec import KatSpec


_MODEL = Model()
_SPEC = KatSpec()

# Supported kat script operators.
BINARY_OPERATORS = _SPEC.binary_operators
UNARY_OPERATORS = _SPEC.unary_operators
CONSTANTS = _SPEC.constants
EXPRESSION_FUNCTIONS = _SPEC.expression_functions


def dedent_multiline(text):
    """Dedent multiline text, stripping preceding and succeeding newlines.

    This is useful for specifying multiline strings in tests without having to
    compensate for the indentation.
    """
    return dedent(text).strip()



"""Simple beam tracing tests - i.e. single cavities, no beam splitters."""

from collections import OrderedDict
import pytest

from finesse.cymath.gaussbeam import transform_q
from finesse.components import Cavity, Laser, Mirror
from finesse.exceptions import BeamTraceException
from finesse.gaussian import BeamParam


def test_laser_to_mirror_trace(model):
    """Test beam tracing for a laser -> mirror configuration with a beam parameter set at the \
    laser."""
    m_roc = 2.5
    L0_q = -0.5 + 1j
    model.chain(Laser("L0"), 1, Mirror("M", Rc=m_roc))

    # First check that we get an exception when no beam parameters are set anywhere.
    with pytest.raises(BeamTraceException):
        model.beam_trace()

    model.L0.p1.o.q = L0_q
    result = model.beam_trace().data_qx

    # TODO: write generic function to compute the target trace results for any given simple config.
    target = OrderedDict()
    target[model.L0.p1.o] = BeamParam(q=L0_q)
    target[model.L0.p1.i] = -target[model.L0.p1.o].conjugate()
    target[model.M.p1.i] = transform_q(
        model.L0.p1.o.space.ABCD(model.L0.p1.o, model.M.p1.i),
        target[model.L0.p1.o],
        1,
        1,
    )
    target[model.M.p1.o] = -target[model.M.p1.i].conjugate()
    target[model.M.p2.o] = transform_q(
        model.M.ABCD(model.M.p1.i, model.M.p2.o), target[model.M.p1.i], 1, 1,
    )
    target[model.M.p2.i] = -target[model.M.p2.o].conjugate()

    for node, q in target.items():
        assert q == result[node]


def test_fabry_perot_no_cav_trace(model):
    """Test beam tracing for a laser -> mirror -> space -> mirror configuration with a beam \
    parameter set at the laser and no cavity object added."""
    m_roc = 2.5
    L0_q = -0.9 + 1.3j
    model.chain(Laser("L0"), 0.5, Mirror("ITM", Rc=-m_roc), 1, Mirror("ETM", Rc=m_roc))
    model.L0.p1.o.q = L0_q
    result = model.beam_trace().data_qx

    # TODO: write generic function to compute the target trace results for any given simple config.
    target = OrderedDict()
    # L0 p1
    target[model.L0.p1.o] = BeamParam(q=L0_q)
    target[model.L0.p1.i] = -target[model.L0.p1.o].conjugate()
    # ITM p1
    target[model.ITM.p1.i] = transform_q(
        model.L0.p1.o.space.ABCD(model.L0.p1.o, model.ITM.p1.i),
        target[model.L0.p1.o],
        1,
        1,
    )
    target[model.ITM.p1.o] = -target[model.ITM.p1.i].conjugate()
    # ITM p2
    target[model.ITM.p2.o] = transform_q(
        model.ITM.ABCD(model.ITM.p1.i, model.ITM.p2.o), target[model.ITM.p1.i], 1, 1,
    )
    target[model.ITM.p2.i] = -target[model.ITM.p2.o].conjugate()
    # ETM p1
    target[model.ETM.p1.i] = transform_q(
        model.ITM.p2.o.space.ABCD(model.ITM.p2.o, model.ETM.p1.i),
        target[model.ITM.p2.o],
        1,
        1,
    )
    target[model.ETM.p1.o] = -target[model.ETM.p1.i].conjugate()
    # ETM p2
    target[model.ETM.p2.o] = transform_q(
        model.ETM.ABCD(model.ETM.p1.i, model.ETM.p2.o), target[model.ETM.p1.i], 1, 1,
    )
    target[model.ETM.p2.i] = -target[model.ETM.p2.o].conjugate()

    for node, q in target.items():
        assert q == result[node]


def test_fabry_perot_cav_trace(model):
    """Test beam tracing for a laser -> mirror -> space -> mirror configuration with a cavity
    object added."""
    m_roc = 2.5
    model.chain(Laser("L0"), 0.5, Mirror("ITM", Rc=-m_roc), 1, Mirror("ETM", Rc=m_roc))
    model.add(Cavity("FPC", model.ITM.p2.o, model.ITM.p2.i))
    result = model.beam_trace().data_qx

    # TODO: write generic function to compute the target trace results for any given simple config.
    target = OrderedDict()
    # ITM p2
    target[model.ITM.p2.o] = model.FPC.qx
    target[model.ITM.p2.i] = -target[model.ITM.p2.o].conjugate()
    # ETM p1
    target[model.ETM.p1.i] = transform_q(
        model.ITM.p2.o.space.ABCD(model.ITM.p2.o, model.ETM.p1.i),
        target[model.ITM.p2.o],
        1,
        1,
    )
    target[model.ETM.p1.o] = -target[model.ETM.p1.i].conjugate()
    # ETM p2
    target[model.ETM.p2.o] = transform_q(
        model.ETM.ABCD(model.ETM.p1.i, model.ETM.p2.o), target[model.ETM.p1.i], 1, 1,
    )
    target[model.ETM.p2.i] = -target[model.ETM.p2.o].conjugate()
    # ITM p1
    target[model.ITM.p1.o] = transform_q(
        model.ITM.ABCD(model.ITM.p2.i, model.ITM.p1.o), target[model.ITM.p2.i], 1, 1,
    )
    target[model.ITM.p1.i] = -target[model.ITM.p1.o].conjugate()
    # L0 p1
    target[model.L0.p1.i] = transform_q(
        model.ITM.p1.o.space.ABCD(model.ITM.p1.o, model.L0.p1.i),
        target[model.ITM.p1.o],
        1,
        1,
    )
    target[model.L0.p1.o] = -target[model.L0.p1.i].conjugate()

    for node, q in target.items():
        assert q == result[node]


def test_fabry_perot_cav_and_manual_trace(model):
    """Test beam tracing for a laser -> mirror -> space -> mirror configuration with a cavity \
    object added and a beam parameter set at the laser."""
    m_roc = 2.5
    L0_q = -1.1 + 1.6j
    model.chain(Laser("L0"), 0.5, Mirror("ITM", Rc=-m_roc), 1, Mirror("ETM", Rc=m_roc))
    model.L0.p1.o.q = L0_q
    model.add(Cavity("FPC", model.ITM.p2.o, model.ITM.p2.i))
    result = model.beam_trace().data_qx

    # TODO: write generic function to compute the target trace results for any given simple config.
    target = OrderedDict()
    # L0 p1
    target[model.L0.p1.o] = BeamParam(q=L0_q)
    target[model.L0.p1.i] = -target[model.L0.p1.o].conjugate()
    # ITM p2
    target[model.ITM.p2.o] = model.FPC.qx
    target[model.ITM.p2.i] = -target[model.ITM.p2.o].conjugate()
    # ETM p1
    target[model.ETM.p1.i] = transform_q(
        model.ITM.p2.o.space.ABCD(model.ITM.p2.o, model.ETM.p1.i),
        target[model.ITM.p2.o],
        1,
        1,
    )
    target[model.ETM.p1.o] = -target[model.ETM.p1.i].conjugate()
    # ETM p2
    target[model.ETM.p2.o] = transform_q(
        model.ETM.ABCD(model.ETM.p1.i, model.ETM.p2.o), target[model.ETM.p1.i], 1, 1,
    )
    target[model.ETM.p2.i] = -target[model.ETM.p2.o].conjugate()
    # ITM p1
    target[model.ITM.p1.i] = transform_q(
        model.L0.p1.o.space.ABCD(model.L0.p1.o, model.ITM.p1.i),
        target[model.L0.p1.o],
        1,
        1,
    )
    target[model.ITM.p1.o] = -target[model.ITM.p1.i].conjugate()

    for node, q in target.items():
        assert q == result[node]



from functools import partial
import pytest


@pytest.fixture
def compiler(monkeypatch, compiler, model):
    monkeypatch.setattr(compiler, "compile", partial(compiler.compile, model=model))
    return compiler


@pytest.fixture
def model_matcher(compiler, model):
    def matcher(script, elements):
        compiler.compile(script)
        # Just check names for simplicity.
        assert set(
            element.name
            for element in model.elements.values()
            if element.name != "fsig"
        ) == set(element.name for element in elements)

    return matcher



"""Laser-type optical components for producing beams."""
import logging
import math
import types
import numpy as np

from finesse.cymath.complex import crotate
from finesse.parameter import float_parameter

from finesse.components.general import (
    Connector,
    FrequencyGenerator,
    NoiseType,
    DOFDefinition,
)
from finesse.components.node import NodeType, NodeDirection

LOGGER = logging.getLogger(__name__)


@float_parameter("P", "Power", units="W")
@float_parameter("phase", "Phase", units="degrees")
@float_parameter("f", "Frequency", units="Hz")
class Laser(Connector, FrequencyGenerator):
    """Represents a laser producing a beam with associated properties such as power and
    frequency.

    Parameters
    ----------
    name : str
        Name of the newly created laser.

    P : float, optional
        Power of the laser (in Watts), defaults to 1 W.

    f : float or :class:`.Frequency`, optional
        Frequency-offset of the laser from the default (in Hz) or
        :class:`.Frequency` object. Defaults to 0 Hz offset.

    phase : float, optional
        Phase-offset of the laser from the default, defaults to zero.
    """

    _DEFAULT_POWER_COEFFS = {(0, 0): (1.0, 0.0)}

    def __init__(self, name, P=1, f=0, phase=0):
        Connector.__init__(self, name)
        FrequencyGenerator.__init__(self)

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        # Modulation inputs
        self._add_port("amp", NodeType.ELECTRICAL)
        self.amp._add_node("i", NodeDirection.INPUT)
        self._add_port("phs", NodeType.ELECTRICAL)
        self.phs._add_node("i", NodeDirection.INPUT)
        self._add_port("frq", NodeType.ELECTRICAL)
        self.frq._add_node("i", NodeDirection.INPUT)

        self._register_node_coupling("SIGAMP_P1o", self.amp.i, self.p1.o)
        self._register_node_coupling("SIGPHS_P1o", self.phs.i, self.p1.o)
        self._register_node_coupling("SIGFRQ_P1o", self.frq.i, self.p1.o)

        self.f = f
        self.P = P
        self.phase = phase
        self.__power_coeffs = self._DEFAULT_POWER_COEFFS.copy()

        # Define typical degrees of freedom for this component
        self.dofs = types.SimpleNamespace()
        self.dofs.amp = DOFDefinition(self.P, self.amp.i, 1)
        self.dofs.phs = DOFDefinition(self.phase, self.phs.i, 1)
        self.dofs.frq = DOFDefinition(self.f, self.frq.i, 1)

    def _source_frequencies(self):
        return [self.f.ref]

    @property
    def power_coeffs(self):
        """The relative power factors and phase offsets for each HGnm mode.

        :getter: Returns the mode factors and phase offsets as a dict with
                 the mode indices as keys. Read-only.
        """
        return self.__power_coeffs.copy()

    @property
    def non_default_power_coeffs(self):
        """The power factors and phase offsets excluding the default.

        :getter: Returns the non-default mode factors and phase offsets as a
                 dict with the mode indices as keys. Read-only.
        """
        coeffs = self.power_coeffs
        default_coeffs = self._DEFAULT_POWER_COEFFS
        if default_coeffs.items() <= coeffs.items():
            # The current coefficients contain the defaults, so remove them.
            for key in default_coeffs:
                coeffs.pop(key)
        return coeffs

    def tem(self, n, m, factor, phase=0.0):
        """Distributes power into the mode HGnm.

        .. note::
            This does not change the total power of the laser, rather, it
            re-distributes this power into / out of the specified mode.

        Parameters
        ----------
        n, m : int
            Mode indices.

        factor : number
            Relative power factor, modes with equal `factor` will
            have equivalent power distributed to them.

        phase : number, optional; default = 0
            Phase offset for the field, in degrees.
        """
        self.__power_coeffs[(n, m)] = factor, phase

    def __find_src_freq(self, sim):
        # if it's tunable we want to look for the symbol that is just this
        # lasers frequency, as it will be changing
        for f in sim.optical_frequencies.frequencies:
            if not self.f.is_changing:
                # Don't match changing frequency bins if ours won't match
                if not f.symbol.is_changing and (
                    f.f == self.f.value  # match potential param refs
                    or f.f == float(self.f.value)  # match numeric values
                ):
                    # If nothing is changing then we can just match freq values
                    return f
            else:
                # If our frequency is changing then we have to have a frequency bin that
                # matches our symbol
                if f.symbol == self.f.ref:
                    return f  # Simple case
        return None

    def _get_workspace(self, sim):
        from finesse.components.modal.laser import (
            laser_carrier_fill_rhs,
            laser_fill_signal,
            laser_fill_qnoise,
            LaserWorkspace,
            laser_set_gouy,
        )

        ws = LaserWorkspace(self, sim, True)
        ws.node_car_id = sim.carrier.node_id(self.p1.o)
        ws.fsrc_car_idx = -1

        ws.set_gouy_function(laser_set_gouy)
        # Carrier just fills RHS
        ws.carrier.set_fill_rhs_fn(laser_carrier_fill_rhs)

        fsrc = self.__find_src_freq(sim.carrier)
        # Didn't find a Frequency bin for this laser in carrier simulation
        if fsrc is None:
            raise Exception(f"Could not find a frequency bin at {self.f} for {self}")
        ws.fsrc_car_idx = fsrc.index

        if sim.is_modal:
            scaling = 0
            ws.power_coeffs = np.zeros(sim.model_data.num_HOMs, dtype=np.complex128)
            coeffs = self.power_coeffs
            for i in range(sim.model_data.num_HOMs):
                n = sim.model_data.homs_view[i][0]
                m = sim.model_data.homs_view[i][1]

                try:
                    factor, phase = coeffs.pop((n, m))
                except KeyError:
                    factor = phase = 0

                ws.power_coeffs[i] = crotate(
                    complex(math.sqrt(factor), 0), math.radians(phase)
                )
                scaling += abs(ws.power_coeffs[i]) ** 2

            if not scaling:
                raise RuntimeError(
                    f"No power in any modes of {self.name}! At least one mode "
                    "must have a non-zero power factor applied to it."
                )

            for i in range(sim.model_data.num_HOMs):
                ws.power_coeffs[i] /= np.sqrt(scaling)

            if coeffs:
                LOGGER.warning(
                    "The following modes, included in the coeffs of %s, are not being "
                    "modelled and will be ignored: %s",
                    self.name,
                    list(coeffs.keys()),
                )

        if sim.signal:
            ws.node_sig_id = sim.signal.node_id(self.p1.o)
            # Audio sim requies matrix filling
            # for signal couplings
            ws.signal.add_fill_function(laser_fill_signal, True) # TODO sort out refill flag here
            ws.signal.set_fill_noise_function(NoiseType.QUANTUM, laser_fill_qnoise)
            # Find the sideband frequencies
            sb = tuple(
                (
                    f
                    for f in sim.signal.optical_frequencies.frequencies
                    if f.audio_carrier_index == fsrc.index
                )
            )
            if len(sb) != 2:
                raise Exception(
                    f"Only something other than two audio sidebands {sb} for carrier {fsrc}"
                )
            ws.fcar_sig_sb_idx = (sb[0].index, sb[1].index)

        # if sim.is_modal: self._update_tem_gouy_phases(sim)
        return ws

    def _couples_frequency(self, ws, connection, frequency_in, frequency_out):
        # The only connections we have are signal inputs to optical output
        # And all the inputs should generate any output.
        return True



import pytest
from ....util import dedent_multiline


@pytest.fixture
def compiler(compiler, fake_binop_cls, fake_unop_cls):
    element = lambda a, b: None
    compiler.spec.register_element(
        "fake_element", {"setter": element, "getter": element}
    )
    compiler.spec.register_element(
        "fake_build_last", {"setter": element, "getter": element, "build_last": True},
    )
    analysis = lambda a: None
    compiler.spec.register_analysis(
        "fake_analysis", {"setter": analysis, "getter": analysis}
    )
    compiler.spec.binary_operators["+"] = lambda a, b: fake_binop_cls("+", a, b)
    compiler.spec.binary_operators["-"] = lambda a, b: fake_binop_cls("-", a, b)
    compiler.spec.unary_operators["-"] = lambda a: fake_unop_cls("-", a)

    return compiler


@pytest.mark.parametrize(
    "script,expected",
    (
        pytest.param(
            # el2 is a dependency to el1.
            dedent_multiline(
                """
                fake_element el1 &el2.a
                fake_element el2
                """
            ),
            [".script.1", ".script.0"],
            id="dependency to another element",
        ),
        pytest.param(
            # el1 is a dependency to el2. el2 also contains a self-reference but this edge is not
            # added to the graph and is instead stored separately.
            dedent_multiline(
                """
                fake_element el1 1 2
                fake_element el2 &el1.a 1-&el2.a
                """
            ),
            [".script.0", ".script.1"],
            id="dependency to another element and self",
        ),
        pytest.param(
            # The nested analysis should not appear in the dependency graph.
            dedent_multiline(
                """
                fake_analysis(fake_analysis())
                """
            ),
            [".script.0"],
            id="nested dependency",
        ),
        pytest.param(
            dedent_multiline(
                """
                fake_analysis(param=-5000+&el1.a)
                fake_element el1 1e5
                """
            ),
            [".script.1", ".script.0"],
            id="nested expression with dependency",
        ),
        pytest.param(
            dedent_multiline(
                """
                fake_build_last el1
                fake_element el2 1e5
                """
            ),
            [".script.1", ".script.0"],
            id="directive with `build_last` flag",
        ),
    ),
)
def test_build_order(compiler, script, expected):
    compiler.compile(script)
    build_order = sorted(
        compiler.graph.dependent_argument_nodes(compiler.graph.ROOT_NODE_NAME),
        key=compiler.graph.argument_node_order,
    )
    assert build_order == expected



import pytest
from ....util import dedent_multiline


@pytest.fixture
def compiler(compiler):
    element = lambda a=None, b=None, c=None, d=None: None
    compiler.spec.register_element(
        "fake_element", {"setter": element, "getter": element}
    )
    compiler.spec.keywords = {"E", "F", "G", "H"}
    return compiler


@pytest.mark.parametrize(
    "script,graph_definition",
    (
        pytest.param(
            "fake_element myelement E F G H",
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label ".script"
                        type "ROOT"
                        extra_tokens "[]"
                    ]
                    node [
                        id 1
                        label ".script.0"
                        type "ELEMENT"
                        token "'fake_element'"
                        name_token "'myelement'"
                        extra_tokens "[' ', ' ', ' ', ' ', ' ']"
                    ]
                    node [
                        id 2
                        label ".script.0.0"
                        token "'E'"
                        type "KEYWORD"
                    ]
                    node [
                        id 3
                        label ".script.0.1"
                        token "'F'"
                        type "KEYWORD"
                    ]
                    node [
                        id 4
                        label ".script.0.2"
                        token "'G'"
                        type "KEYWORD"
                    ]
                    node [
                        id 5
                        label ".script.0.3"
                        token "'H'"
                        type "KEYWORD"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 1
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 3
                        target 1
                        type "ARGUMENT"
                        order "1"
                    ]
                    edge [
                        source 4
                        target 1
                        type "ARGUMENT"
                        order "2"
                    ]
                    edge [
                        source 5
                        target 1
                        type "ARGUMENT"
                        order "3"
                    ]
                ]
                """
            ),
            id="element with 4 keywords",
        ),
    ),
)
def test_keywords(assert_graphs_match, script, graph_definition):
    assert_graphs_match(script, graph_definition)



"""Kat script command line interface"""

import click

from .. import __version__, PROGRAM
from ..cli import (
    KatState,
    plot_graph,
    input_file_argument,
    graph_layout_argument,
    graphviz_option,
    verbose_option,
    quiet_option,
    debug_option,
    log_exclude_option,
)
from .exceptions import KatScriptError


BOUND_WIDTH = 18


@click.group(help="Kat script interface.")
@click.version_option(version=__version__, prog_name=PROGRAM)
@click.pass_context
def kat_script(ctx):
    pass


@kat_script.command()
@input_file_argument
@click.option("--raw", is_flag=True, default=False, help="Show raw tokens.")
@verbose_option
@quiet_option
@debug_option
@log_exclude_option
@click.pass_context
def tokenize(ctx, input_file, raw):
    """Tokenize a Finesse script."""

    from .tokenizer import KatTokenizer

    def print_tokens(tokens):
        for token in tokens:
            value = repr(token) if raw else str(token)
            state.print(f"{token.bounds!s:>{BOUND_WIDTH}} {value}")

    state = ctx.ensure_object(KatState)
    state.print_banner(input_file)

    tokenizer = KatTokenizer()

    try:
        tokens = list(tokenizer.tokenize_file(input_file))
    except KatScriptError as e:
        state.print_error(e)

    state.print("Tokens:")
    print_tokens(tokens)


@kat_script.command()
@input_file_argument
@click.option("--raw", is_flag=True, default=False, help="Show raw parsed productions.")
@verbose_option
@quiet_option
@debug_option
@log_exclude_option
@click.pass_context
def parse(ctx, input_file, raw):
    """Parse a Finesse script and display items."""
    state = ctx.ensure_object(KatState)
    state.print_banner(input_file)

    from .parser import KatParser

    parser = KatParser()

    try:
        script = parser.parse_file(input_file)
    except KatScriptError as e:
        state.print_error(e)

    state.print("Statements:")
    if script.arguments:
        for item in script.arguments:
            value = repr(item) if raw else str(item)
            state.print(f"{item.bounds!s:>{BOUND_WIDTH}} {value}")
    else:
        state.print("*none*")
    state.print()
    state.print("Extra tokens:")
    if script.extra:
        for extra in script.extra:
            if raw:
                state.print(repr(extra))
            else:
                state.print(f"{extra.bounds!s:>{BOUND_WIDTH}} {extra}")
    else:
        state.print("*none*")


@kat_script.command()
@input_file_argument
@click.option(
    "--resolve/--no-resolve",
    is_flag=True,
    default=True,
    help="Resolve model references.",
)
@click.option(
    "--build/--no-build", is_flag=True, default=True, help="Build the model.",
)
@click.option("--graph", is_flag=True, default=False, help="Show graph.")
@click.option(
    "--dump-graph", is_flag=True, default=False, help="Dump the graph to GML markup."
)
@click.option(
    "--build-order", is_flag=True, default=False, help="Show only the build graph.",
)
@click.option("--node-labels", is_flag=True, default=True, help="Show node labels.")
@click.option(
    "--node-attr",
    multiple=True,
    help=(
        "Show node attribute. Can be specified multiple times. If not specified, all attributes "
        "are shown."
    ),
)
@click.option(
    "--edge-attr",
    multiple=True,
    help=(
        "Show edge attribute. Can be specified multiple times. If not specified, all attributes "
        "are shown."
    ),
)
@graph_layout_argument
@graphviz_option
@verbose_option
@quiet_option
@debug_option
@log_exclude_option
@click.pass_context
def compile(
    ctx,
    input_file,
    resolve,
    build,
    graph,
    dump_graph,
    build_order,
    node_labels,
    node_attr,
    edge_attr,
    layout,
    graphviz,
):
    """Compile a Finesse script and display the model."""
    from .compiler import KatCompiler

    state = ctx.ensure_object(KatState)
    state.print_banner(input_file)

    compiler = KatCompiler()

    model = compiler.compile_file(input_file, resolve=resolve, build=build)

    if model:
        state.print(model.info())

    if graph:
        view_graph = compiler.build_graph if build_order else compiler.graph

        if dump_graph:
            from finesse.utilities import stringify_graph_gml

            state.print(stringify_graph_gml(view_graph))

        if not node_attr:
            node_attr = True

        if not edge_attr:
            edge_attr = True

        kwargs = {}

        if not graphviz:
            # Apply colours to the nodes.
            kwargs["node_color_key"] = lambda node, data: data["type"]

        plot_graph(
            state,
            view_graph,
            layout,
            node_labels=node_labels,
            node_attrs=node_attr,
            edge_attrs=edge_attr,
            graphviz=graphviz,
            **kwargs,
        )


@kat_script.command()
@input_file_argument
@click.option("--graph", is_flag=True, default=False, help="Show graph.")
@click.option(
    "--dump-graph", is_flag=True, default=False, help="Dump the graph to GML markup."
)
@click.option("--node-labels", is_flag=True, default=True, help="Show node labels.")
@click.option(
    "--node-attr",
    multiple=True,
    help=(
        "Show node attribute. Can be specified multiple times. If not specified, all attributes "
        "are shown."
    ),
)
@click.option(
    "--edge-attr",
    multiple=True,
    help=(
        "Show edge attribute. Can be specified multiple times. If not specified, all attributes "
        "are shown."
    ),
)
@graph_layout_argument
@graphviz_option
@verbose_option
@quiet_option
@debug_option
@log_exclude_option
@click.pass_context
def normalise(
    ctx,
    input_file,
    graph,
    dump_graph,
    node_labels,
    node_attr,
    edge_attr,
    layout,
    graphviz,
):
    """Compile a Finesse script and print normalised kat script."""
    from .compiler import KatCompiler
    from .generator import KatUnbuilder

    state = ctx.ensure_object(KatState)
    state.print_banner(input_file)

    compiler = KatCompiler()

    model = compiler.compile_file(input_file)
    unbuilder = KatUnbuilder()
    script = unbuilder.unbuild(model)

    state.print(script)

    if graph:
        view_graph = unbuilder.graph

        if dump_graph:
            from finesse.utilities import stringify_graph_gml

            state.print(stringify_graph_gml(view_graph))

        if not node_attr:
            node_attr = True

        if not edge_attr:
            edge_attr = True

        kwargs = {}

        if not graphviz:
            # Apply colours to the nodes.
            kwargs["node_color_key"] = lambda node, data: data["type"]

        plot_graph(
            state,
            view_graph,
            layout,
            node_labels=node_labels,
            node_attrs=node_attr,
            edge_attrs=edge_attr,
            graphviz=graphviz,
            **kwargs,
        )


@kat_script.command()
@verbose_option
@quiet_option
@debug_option
@click.pass_context
def shell(ctx):
    """Start an interactive Finesse shell."""
    from ..script.shell import KatShell

    repl = KatShell()
    repl.cmdloop()


if __name__ == "__main__":
    kat_script()



import pytest
import logging
from os import path
import numpy as np
from finesse import Model

# this gets regenetered each
# time which is quite expesive
# TODO optimise this unit test
# if it gets expensive to run
@pytest.fixture
def output_test_obj(noxaxis):
    """ Executed model object """
    ifo = Model()

    ifo.parse_legacy(
        """
l l1 1 0 0 ni
gauss g1 l1 ni 1m 0
s s1 10 ni n1
maxtem 1
tem l1 0 0 0 0
tem l1 1 0 1 0
phase 0
ad ad00 0 0 0 n1
ad ad10 1 0 0 n1
pd pd n1
#beam ccd 0 n1 this causes an error
yaxis re:im
        """
    )
    if noxaxis:
        ifo.parse_legacy("noxaxis")
    else:
        ifo.parse_legacy("xaxis s1 L lin 1 2 5")

    out = ifo.run()
    return out, ifo


@pytest.mark.parametrize("noxaxis", (True, False))
def test_get_legacy_data(output_test_obj):
    out, ifo = output_test_obj
    legacy_data, column_names, plot_type = out.get_legacy_data(ifo)
    assert isinstance(legacy_data, np.ndarray)
    assert isinstance(column_names, list)
    assert isinstance(plot_type, str)


# TODO: refactor this test
@pytest.mark.parametrize("noxaxis", (True,))
def test_write_legacy_data(caplog, tmpdir, output_test_obj):
    out, ifo = output_test_obj
    legacy_data, column_names, plot_type = out.get_legacy_data(ifo)

    # set one equal to None and check it emits warning
    Fout = tmpdir.join("output.out")
    with caplog.at_level(logging.WARNING):
        out.write_legacy_data(
            ifo,
            filename=Fout,
            legacy_data=legacy_data,
            column_names=None,
            plot_type=plot_type,
        )
        assert "recomputed" in caplog.text

    caplog.clear()
    with caplog.at_level(logging.WARNING):
        out.write_legacy_data(
            ifo,
            filename=Fout,
            legacy_data=legacy_data,
            column_names=column_names,
            plot_type=plot_type,
        )
        assert "recomputed" not in caplog.text

    # check file is actually ouput
    assert path.isfile(Fout)



import sys
import logging
import datetime
import textwrap
from itertools import chain
import click
import tabulate
from .. import __version__, PROGRAM
from ..environment import show_tracebacks
from ..utilities import networkx_layouts, graphviz_layouts, add_linenos, ngettext
from ..script.exceptions import KatScriptError
from ..config import log_handler_instance

LOGGER = logging.getLogger(PROGRAM)


def set_verbosity(ctx, option, value):
    """Callback to set stdout verbosity for root CLI command."""
    # If neither -v nor -q is set, value is 0.
    if not value:
        return

    state = ctx.ensure_object(KatState)

    if option.name == "verbose":
        offset = -10 * value
    elif option.name == "quiet":
        offset = 10 * value
    else:
        raise ValueError("Unsupported option.")

    state.verbosity = state.VERBOSITY_DEFAULT + offset


def set_fancy_error_formatting(ctx, option, value):
    if value is None:
        return

    state = ctx.ensure_object(KatState)
    state.fancy_errors = value


def set_debug(ctx, option, value):
    if value is None:
        return

    state = ctx.ensure_object(KatState)
    state.debug = value


def set_log_excludes(ctx, option, value):
    if value is None:
        return

    state = ctx.ensure_object(KatState)
    for exclude in value:
        state.exclude(exclude)


def print_banner(ctx, option, value):
    """Show the Finesse banner and exit."""
    if not value or ctx.resilient_parsing:  # Keep this check as-is!
        return

    state = ctx.ensure_object(KatState)
    state.print_banner(exit_=True)


def parse_path(state, path, legacy=False, **error_kwargs):
    from finesse.script import parse_file, parse_legacy_file

    try:
        if legacy:
            return parse_legacy_file(path)
        else:
            return parse_file(path)
    except Exception as error:
        state.print_error(error, "PARSING ERROR:", **error_kwargs)


def plot_graph(state, *args, **kwargs):
    from finesse.plotting.graph import plot_graph as finesse_plot_graph

    finesse_plot_graph(*args, **kwargs)


def list_graph_layouts(ctx, option, value):
    """Callback to list available graph layouts."""
    from ..environment import has_pygraphviz

    if not value:
        return

    state = ctx.ensure_object(KatState)

    state.print("Available layouts:", fg="green")

    nx_layouts = networkx_layouts()
    gv_layouts = graphviz_layouts()
    layouts = sorted(chain(nx_layouts, gv_layouts))
    gv_suffix = click.style("*", fg="yellow")

    for layout in layouts:
        suffix = gv_suffix if layout in gv_layouts else ""
        state.print(f"- {layout}{suffix}")

    state.print()
    if has_pygraphviz():
        state.print(
            "*This layout can also be plotted directly in graphviz using the "
            "--graphviz flag",
            fg="yellow",
        )
    else:
        state.print("Install pygraphviz/graphviz for more layouts", fg="yellow")

    sys.exit()


def _fancy_format_kat_error(exception):
    """Format each marked error in red."""
    assert isinstance(exception, KatScriptError)

    error_msg = exception.rubric()

    msg = f"{error_msg}\n"
    errorlines = []
    linenos = []
    errorlinenos = set()
    for lineno, linechunks in exception.chunkify().items():
        line = ""
        for bounds, is_error in linechunks:
            chunktxt = exception.container.script(bounds)

            if is_error:
                errorlinenos.add(lineno)

                if len(chunktxt) == chunktxt.count(" "):
                    # Error chunk is just spaces. Replace with interpunct.
                    chunktxt = click.style("·" * len(chunktxt), fg="red")
                elif chunktxt == "\n":
                    # Error is a newline. Replace with carriage return symbol. Also add the newline
                    # anyway without colouring it red.
                    chunktxt = click.style("↵", fg="red") + chunktxt

                line += click.style(chunktxt, fg="red")
            else:
                line += chunktxt

        # Get rid of trailing newline.
        try:
            line = line.splitlines()[0]
        except IndexError:
            # Line is empty.
            pass

        linenos.append(lineno)
        errorlines.append(line)

    # Add line numbers.
    numberedlines = add_linenos(linenos, errorlines)

    # Add message for missing lines.
    finallines = []
    lastlineno = None
    for lineno, line in zip(linenos, numberedlines):
        if lastlineno:
            diff = lineno - lastlineno
            if diff > 1:
                missingmsg = ngettext(diff - 1, "%d missing line", "%d missing lines")
                finallines.append(f"   *** {missingmsg} ***")
        prefix = "-->" if lineno in errorlinenos else "   "
        finallines.append(prefix + line)
        lastlineno = lineno

    return msg + "\n".join(finallines)


input_file_argument = click.argument("input_file", type=click.File("r"))
output_file_argument = click.argument("output_file", type=click.File("w"), default="-")
plot_option = click.option(
    "--plot/--no-plot",
    default=True,
    show_default=True,
    help="Display results as figure.",
)
trace_option = click.option(
    "--trace/--no-trace",
    default=False,
    show_default=True,
    help="Displays the results of a beam trace of the model.",
)
graphviz_option = click.option(
    "--graphviz",
    is_flag=True,
    default=False,
    show_default=True,
    help="Generate layout and display using Graphviz.",
)
list_graph_layouts_option = click.option(
    "--list-graph-layouts",
    callback=list_graph_layouts,
    is_flag=True,
    default=False,
    expose_value=False,
    is_eager=True,
    help="Show available graph layouts and exit.",
)
graph_layout_argument = click.option(
    "--layout", type=str, default="spring", help="Graph layout algorithm to use.",
)
network_type_argument = click.option(
    "-t",
    "--type",
    "network_type",
    type=click.Choice(("full", "components", "optical")),
    default="full",
    help="Network to plot.",
)
table_format_argument = click.option(
    "--table-format",
    type=click.Choice(tabulate.tabulate_formats),
    default="fancy_grid",
    help="Table format to use.",
)
legacy_option = click.option(
    "--legacy",
    is_flag=True,
    default=False,
    show_default=True,
    help="Specify that the input file uses Finesse 2 syntax.",
)
verbose_option = click.option(
    "-v",
    "--verbose",
    count=True,
    callback=set_verbosity,
    expose_value=False,
    is_eager=True,
    help="Increase verbosity of log output (can be specified multiple times).",
)
quiet_option = click.option(
    "-q",
    "--quiet",
    count=True,
    callback=set_verbosity,
    expose_value=False,
    help="Decrease verbosity of log output (can be specified multiple times).",
)
fancy_error_option = click.option(
    "--fancy-errors/--no-fancy-errors",
    is_flag=True,
    default=True,
    callback=set_fancy_error_formatting,
    expose_value=False,
    show_default=True,
    help="Highlight script error locations in red rather than marking them on the following line.",
)
debug_option = click.option(
    "--debug",
    is_flag=True,
    default=False,
    callback=set_debug,
    expose_value=False,
    show_default=True,
    help="Enable debug mode (print tracebacks).",
)
log_exclude_option = click.option(
    "--log-exclude",
    multiple=True,
    callback=set_log_excludes,
    expose_value=False,
    help="Ignore log records from a particular logger (wildcards allowed).",
)


class ClickLogColorFormatter(logging.Formatter):
    """Stdout log formatter with colors."""

    # Log prefix format.
    # NOTE: this is NOT a :py:class:`logging.Formatter` compatible format string, but rather a
    # format specific to this class.
    PREFIX_FORMAT = "{short_name:>16s} [{record.levelname:>8s}]:"

    STYLES = {
        "critical": {"fg": "red"},
        "error": {"fg": "red"},
        "warning": {"fg": "yellow"},
        "info": {"fg": "green"},
        "debug": {"fg": "blue"},
    }

    def format(self, record):
        if record.exc_info:
            # Skip formatting of exception info.
            return super().format(record)

        level = record.levelname.casefold()
        msg = record.getMessage()

        if level in self.STYLES:
            # Remove "finesse." prefix.
            short_name = (
                record.name[8:] if record.name.startswith("finesse.") else record.name
            )

            prefix = click.style(
                self.PREFIX_FORMAT.format(short_name=short_name, record=record),
                **self.STYLES[level],
            )
            msg = "\n".join(f"{prefix} {line}" for line in msg.splitlines())

        return msg


class KatState:
    """Shared state for all CLI subcommands.

    This object gets built by Click when the CLI is called and encapsulates global
    settings for use by individual commands/groups.
    """

    VERBOSITY_DEFAULT = logging.WARNING

    def __init__(self):
        # Fancy error flag. This is NOT the default value for the CLI - it is set on by
        # default via the @fancy_error_option decorators.
        self.fancy_errors = False

        # Set up the Finesse logger.
        handler = log_handler_instance()
        handler.setFormatter(ClickLogColorFormatter())
        handler.setStream(click.get_text_stream("stderr"))
        LOGGER.addHandler(handler)

        # Set *default* verbosity.
        # NOTE: the user's requested verbosity is set by :func:`set_verbosity`.
        self.verbosity = self.VERBOSITY_DEFAULT

    @property
    def verbosity(self):
        """Log verbosity on stdout."""
        return LOGGER.getEffectiveLevel()

    @verbosity.setter
    def verbosity(self, verbosity):
        verbosity = min(logging.CRITICAL, max(logging.DEBUG, verbosity))
        LOGGER.setLevel(verbosity)

    @property
    def verbose(self):
        """Verbose output enabled.

        Returns True if the verbosity is enough for INFO or DEBUG messages to be
        displayed.
        """
        return self.verbosity <= self.VERBOSITY_DEFAULT

    @property
    def debug(self):
        return show_tracebacks()

    @debug.setter
    def debug(self, value):
        show_tracebacks(value)

    def exclude(self, pattern):
        """Exclude records from loggers with names matching the specified pattern."""
        log_handler_instance().exclude(pattern)

    def print(self, text="", indent=0, error=False, exit_=False, exit_code=0, **kwargs):
        text = str(text)

        if indent > 0:
            text = textwrap.indent(text, " " * 4)

        click.secho(text, err=error, **kwargs)

        if exit_:
            self.exit(exit_code)

    def print_error(
        self, exception_or_msg, title=None, exit_=True, exit_code=1, **kwargs
    ):
        if title:
            self.print(title, error=True, fg="red", bold=True)

        if self.debug and isinstance(exception_or_msg, BaseException):
            # Print the full traceback.
            from traceback import format_exception

            # Only print the traceback part of the exception, because the message is printed below.
            # FIXME: the `etype` is not used here, and the second parameter set to None results in
            # "NoneType: None" being printed by `TracebackException` after the traceback...
            etype, _, tb = sys.exc_info()
            trace = "".join(format_exception(etype, None, tb))
            self.print(trace, indent=1)

        if self.fancy_errors and isinstance(exception_or_msg, KatScriptError):
            msg = _fancy_format_kat_error(exception_or_msg)
            fg = None
        else:
            # Format the whole error as red.
            msg = str(exception_or_msg)
            fg = "red"

        self.print(msg, error=True, fg=fg, **kwargs)

        if exit_:
            self.exit(exit_code)

    def print_banner(self, kat_file=None, **kwargs):
        """Print the Finesse banner."""
        if not self.verbose:
            return

        input_file = f"Input file: {kat_file.name}" if kat_file else ""
        timenow = datetime.datetime.now().strftime("%c").strip()
        timestr = f"{timenow:>60}"
        banner = rf"""
            ------------------------------------------------------------------------
                                 FINESSE {__version__}
                    o_.-=.       Frequency domain INterferomEter Simulation SoftwarE
                    (\'".\|                         http://www.gwoptics.org/finesse/
                    .>' (_--.
                _=/d   ,^\       {input_file}
                ~~ \)-'   '
                / |
                '  '    {timestr}
            ------------------------------------------------------------------------
            """

        self.print(textwrap.dedent(banner).strip(), **kwargs)

    def exit(self, code=0):
        """Stop execution."""
        sys.exit(code)



"""Tests for locking a Fabry-Perot cavity by applying feedback signal to ITM detuning when \
scanning ETM detuning."""

import pytest
import finesse
from finesse.analysis.actions import RunLocks, Sweep

finesse.plotting.init()


@pytest.fixture(scope="module")
def fabry_perot_with_length_lock():
    """Fabry-Perot model and analysis results."""
    kat = finesse.Model()
    kat.parse(
        """
        l L0 P=1
        s s0 L0.p1 EOM1.p1
        mod EOM1 (
            f=100M
            midx=0.1
            order=1
            mod_type=pm
        )
        s s1 EOM1.p2 ITM.p1

        m ITM R=0.99 T=0.01
        s sCAV ITM.p2 ETM.p1 L=1
        m ETM R=1 T=0 phi=0

        pd1 REFL_I ITM.p2.o &EOM1.f 0

        lock lock_length REFL_I ETM.phi -1 1e-5

        xaxis ETM.phi lin -1 1 100
        """
    )
    return kat

@pytest.mark.skip(reason="Fix https://git.ligo.org/finesse/finesse3/-/issues/236")
def test_xaxis_lock(fabry_perot_with_length_lock):
    kat = fabry_perot_with_length_lock.deepcopy()
    sol = kat.run()
    # Check error signal is kept within accuracy limits
    assert all(sol[kat.lock_length.error_signal.name] <= kat.lock_length.accuracy)

@pytest.mark.skip(reason="Fix https://git.ligo.org/finesse/finesse3/-/issues/236")
def test_pre_step_action(fabry_perot_with_length_lock):
    kat = fabry_perot_with_length_lock.deepcopy()
    lock = RunLocks(kat.lock_length)
    step = Sweep("axis", "ITM.phi", (0.0, 1.0, 2.0), False, pre_step=lock)
    step.run(kat)
    # If lock works then ETM phi tracks ITM phi
    assert abs(kat.ITM.phi - kat.ETM.phi) <= kat.lock_length.accuracy

@pytest.mark.skip(reason="Fix https://git.ligo.org/finesse/finesse3/-/issues/236")
def test_lock_action(fabry_perot_with_length_lock):
    """Checks lock fixes initial offset and lock running in serial.
    Mirror should be brought back to the zero position"""
    kat = fabry_perot_with_length_lock.deepcopy()
    kat.ETM.phi = 0.1
    kat.lock_length.accuracy = 1e-15
    lock = RunLocks(kat.lock_length)
    lock.run(kat)
    assert abs(kat.ETM.phi.value) <= kat.lock_length.accuracy * 10


@pytest.fixture(scope="module")
def lock_action_parsed():
    kat = finesse.Model()
    kat.parse(
        """
        l L0 P=1
        s s0 L0.p1 EOM1.p1
        mod EOM1 (
            f=100M
            midx=0.1
            order=1
            mod_type=pm
        )
        s s1 EOM1.p2 ITM.p1

        m ITM R=0.99 T=0.01
        s sCAV ITM.p2 ETM.p1 L=1
        m ETM R=1 T=0 phi=0.1

        pd1 REFL_I ITM.p2.o &EOM1.f 0

        lock lock_length REFL_I ETM.phi -1 1e-15

        run_locks lock_length
        """
    )
    return kat

@pytest.mark.skip(reason="Fix https://git.ligo.org/finesse/finesse3/-/issues/236")
def test_lock_action_parsed(lock_action_parsed):
    lock_action_parsed.run()
    assert (
        abs(lock_action_parsed.ETM.phi.value)
        <= lock_action_parsed.lock_length.accuracy * 10
    )



"""
Contains the class :class:`.Nothing` which represents an empty/null
point in a configuration.
"""

import numpy as np

from finesse.components.general import Connector, InteractionType
from finesse.components.node import NodeType, NodeDirection
from finesse.components.workspace import ConnectorWorkspace, Connections


class NothingWorkspace(ConnectorWorkspace):
    def __init__(self, owner, sim):
        super().__init__(owner, sim, False, False)


class Nothing(Connector):
    """Represents an empty point in the interferometer configuration.

    `Nothing` is just some point in space that can be connected to. For example, you can use this to
    propagate a beam from some component out to some arbitrary point to make a measurement at. You
    can also use this to split spaces up, if you wanted to measure something inbetween two
    components. It can also be used to replace a component, for example if you want to remove a lens
    or a mirror in some beam path.

    Parameters
    ----------
    name : str
        Name of newly created nothing.
    """

    def __init__(self, name):
        super().__init__(name)

        # Here we register that fact this component will
        # want to have some ports and nodes as well as
        # the couplings between them
        self._add_port("p1", NodeType.OPTICAL)  # front
        self._add_port("p2", NodeType.OPTICAL)  # back
        # input and output optical fields at port 1 (Front face)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)
        # input and output optical fields at port 2 (Back face)
        self.p2._add_node("i", NodeDirection.INPUT)
        self.p2._add_node("o", NodeDirection.OUTPUT)
        # Optic to optic couplings
        self._register_node_coupling(
            "P1i_P2o",
            self.p1.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
        )

    def _get_workspace(self, sim):
        ws = NothingWorkspace(self, sim)
        ws.I = np.eye(sim.model_data.num_HOMs, dtype=np.complex128)
        ws.carrier.add_fill_function(self._fill_carrier, False)
        ws.signal.add_fill_function(self._fill_signal, False)
        return ws

    def _fill_optical_matrix(self, ws, matrix, connections):
        for freq in matrix.optical_frequencies.frequencies:
            with matrix.component_edge_fill3(
                ws.owner_id, connections.P1i_P2o_idx, freq.index, freq.index,
            ) as mat:
                mat[:] = ws.I

            with matrix.component_edge_fill3(
                ws.owner_id, connections.P2i_P1o_idx, freq.index, freq.index,
            ) as mat:
                mat[:] = ws.I

    def _fill_carrier(self, ws):
        self._fill_optical_matrix(ws, ws.sim.carrier, ws.carrier.connections)

    def _fill_signal(self, ws):
        self._fill_optical_matrix(ws, ws.sim.signal, ws.signal.connections)

